﻿; g_spot 02449x ejn (blueznl)
; gadget / object resizing kit
; bluez@xs4all.nl
;
; - resize and reposition gadgets
; - draw 2d elements on an image
;
#gs_version = "02449x"
#gs_copyright = "c2005..2010 ejn"
;
Declare gs_rule(plane.i,rule.i,objecttype.i,objectnr.i,referencetype.i=-1,referencenr.i=-1,modifier1.q=0,modifier2.q=0)


; *** generic

Procedure gs_history()                                                                ; all changes thus far
  ;
  ; to do
  ;
  ; - add new rule gs_plane to add planes through rules
  ; - add new rule gs_limit (to limit position and size to max and min values)
  ; - fill in last few defined but still missing rules
  ; - add new rules gs_spreadh and gs_spreadv
  ;
  ; current
  ;
  ; - improved documentation (text in file, and seperate excel worksheet)
  ; - added gs_panmouse() gs_worldx() gs_worldy()
  ; - added planes, see gs_plane() etc.
  ; - added multiple viewpanes, ports, and associated scales
  ; - added new element gs_marker()
  ; - gs_box() did not support planes, fixed
  ; - added shift support to some more commands such as gs_zoomall()
  ; - fixed some typos in comments :-)
  ;
  ; 02405x
  ;
  ; - added the possibility to shift all elements with x/y per plane
  ; - zooming, panning, viewpane vs. viewport, and more bugfixes
  ; - flattened to use it as an include (need to turn codecaddy into a universal preprocessor, sigh)
  ; - improved structure viewer
  ; - added gs_setgadgetitem
  ; - changed sequence of gs_rule() parameters
  ; - added planes (sort of layers)
  ;
  ; 01319x
  ;
  ; - added gs_internalsviewer() to aid in debugging
  ; - added shadow elements that show a placeholder for a gadget
  ; - multiple 2d elements now working
  ; - multiple bug fixes
  ; - rules now affect images and box elements
  ; - added (some) runaway protection
  ; - added some code to simplify a potential :-) selectmapelement() command
  ; - minor speedup by first checking for existence of gadget and / or window
  ;
  ; 00001x
  ;
  ; - brand new start :-)
  ;
  ;
EndProcedure

Procedure gs_notes()                                                                  ; documentation
  ;
  ; 1. Welcome :-)
  ; --------------
  ;
  ; This include file is going to be part of the upcoming 'sweetspot'... still a secret project :-)
  ;
  ; This include file contains a number of functions, related to:
  ;
  ; - rule based gadget placement and resizing
  ; - 2d library for line drawing etc. (to be expanded upon)
  ;
  ;
  ; 2. Objects.
  ; -----------
  ;
  ; g_spot can manipulate objects such as windows, gadgets, planes, images, and elements.
  ;
  ; When refering to an object you need to uniquely identify the object using a combination of the type of the
  ; object, and its number.
  ;
  ; 'Rules' (see 3) often use two objects: the object affected, and (optional) an object refered to, called
  ; 'reference'.
  ;
  ;    object type     note
  ;    --------------  ----------------------------------------------------------------------------
  ;    #gs_none        no reference, often assumed to be the same as null
  ;    #gs_null        nothing, no reference
  ;    #gs_plane       a plane
  ;    #gs_window      a purebasic window
  ;    #gs_gadget      a purebasic gadget
  ;    #gs_image       a purebasic image
  ;    #gs_element     a 2d element
  ;
  ;
  ; 3. Rules.
  ; ---------
  ;
  ; g_spot allows to add a rule for purebasic gadgets, windows and images. It can detect changes in any of these
  ; elements and resize / reposition other elements depending on rules given by the programmer.
  ;
  ; The programmer does not need any other functionality of g_spot to use this.
  ;
  ;
  ; 3.1 Defining a rule.
  ; --------------------
  ;
  ; Each object (window, gadget, element, image) can be added to the ruleset using the same command:
  ;
  ;   gs_rule( <plane> , <rule> , <objecttype> , <objectnr> , <reftype> , <refnr> [ , <mod1> [ , <mod2> ]] )
  ;
  ; Object and reference are assumed to be on the same plane (more about planes later).
  ;
  ; gs_apply() identifies all rules that can be used (rules that apply or refer to objects which do not
  ; exist are marked as not in use). After gs_apply() the hash table gs_objects() will contain a list of all
  ; recognized objects.
  ;
  ; <plane> can be used to tie a rule to a specific plane, although the elements the rule refers to do not
  ; have to be part of that plane.
  ;
  ; <mod1> and <mod2> are two additional modifiers / optional parameters. Their function depends on the
  ; rule used, but typically they provide an offset x, y, width, height, limit, or a combination thereof.
  ;
  ;
  ; 3.2 Applying rules.
  ; -------------------
  ;
  ; Add the following command to the event loop, and execute it whenever a #pb_event_move or #pb_event_size
  ; event takes place:
  ;
  ;   [ result = ] gs_apply( [<plane>] )
  ;
  ; gs_apply() will resize and / or move all window and gadget objects.
  ;
  ; If any objects were affected, gs_apply() will return true. This can be used to start a (re)draw action
  ; for non-gadget non-window objects, for example all elements:
  ;
  ;   If gs_apply()
  ;     gs_drawallelements()
  ;   EndIf
  ;
  ; See section 5 for more on rule parameters.
  ;
  ;
  ; 3.3 Rule parameters.
  ; --------------------
  ;
  ; Several rules allow parameters, these (may) differ from to rule. The basic command has the same format:
  ;
  ;   gs_rule( <plane> , <rule> , <objecttype> , <objectnr> , <reftype> , <refnr> [ , <mod1> [ , <mod2> ]] )
  ;
  ;
  ; 3.4 Basic field usage.
  ; ----------------------
  ;
  ; Most rules use the fields like this:
  ;
  ;   field         note
  ;   ------------  ------------------------------------------------------------------------------------------------
  ;   <plane>       the rule will be 'tied' to a plane, so selective applying of rules (using
  ;                 gs_apply(<planenr>) is possble)
  ;   <rule>        the rule
  ;   <objecttype>  type of object that is affected by the rule
  ;                 can be #gs_plane, #gs_window, #gs_gadget, #gs_image
  ;   <objectnr>    number of object that is affected by the rule
  ;                 the combination of <objecttype> + <objectnr> is unique
  ;   <reftype>     type of reference that is refered to by the rule
  ;                 can be #gs_plane, #gs_window, #gs_gadget, #gs_image, #gs_null, #gs_none
  ;   <refnr>       number of object that is refered to
  ;   <mod1>        optional modifier, used to set specific coordinates, offsets, limits, or spacing
  ;   <mod2>        similar to mod1
  ;
  ;
  ; 3.5 Fields per rule.
  ; --------------------
  ;
  ;   standard rules                   notes
  ;   -------------------------------  ----------------------------------------------------------------------------
  ;   #gs_none, #gs_null               mod1 and mod2 are not used
  ;   #gs_setxy                        mod1 and mod2 can be used to specify x and y coordinate or offset
  ;                                    when not refering to an object mod1,mod2 specify the coordinate
  ;                                    when refering to an object mod1,mod2 are added to the object's coordinate
  ;   #gs_setx, #gs_sety               mod1 specifies addition (offset or coordinate when not using a reference)
  ;                                    mod2 specifies minimal horizontal or vertical position of the object
  ;   #gs_setwidth, #gs_setheight      as above, mod2 specifies minimal width or height
  ;   #gs_setsize                      specifiy size, or use size of specified object and add mod1,mod2 to size
  ;   #gs_use                          use x, y, width and height of referenced object
  ;   #gs_stackup, #gs_stackdown       place the object above or below the reference, change the size to match
  ;                                    the reference object's width, mod1 specifies a distance between both objects
  ;   #gs_stackleft, #gs_stackright    place the object left or right of reference, change size, mod1 specifies
  ;                                    additional space between the object and the reference
  ;   #gs_alignleft                    alias for setx
  ;   #gs_alignright                   align right side of object and reference
  ;                                    (x2 object = x2 reference + modifier1)
  ;   #gs_aligntop                     alias for sety, align top side of reference and object
  ;   #gs_alignbottom                  align bottom side of object with bottom side of reference
  ;                                    (x2 object = x3 reference + modifier)
  ;   #gs_leftof, #gs_rightof          place object left ie. right of reference, mod1 specifies additional spacing
  ;   #gs_above, #gs_below             place object above ie. below reference, mod1 specifies additional spacing
  ;   #gs_stretchleft                  resize the object, the left side will be stretched to match the right side
  ;                                    of the reference object, mod1 specifies how much space is left between
  ;                                    object and reference, mod2 specifies the minimal width of the object
  ;   #gs_stretchright                 similar to above
  ;   #gs_stretchup, #gs_stretchdown   similar to above
  ;   #gs_x1x1, #gs_x1x2, #gs_x2x1     alias for setx, alias for rightof, alias for leftof
  ;   #gs_x2x2, #gs_y1y1, #gs_y1y2     alias for alignright, alias for sety, alias for below
  ;   #gs_y2y1, #gs_y2y2               alias for above, alias for alignbottom
  ;   #gs_stretchx1x2, #gs_stretchx2x1 alias for stretchleft, alias for stretchright
  ;   #gs_stretchy1y2, #gs_stretchy2y1 alias for stretchup, alias for stretchdown
  ;
  ;   parent related resizing          notes
  ;   -------------------------------  ----------------------------------------------------------------------------
  ;   #gs_stretchparentleft            stretch the left side of the object towards the left side of the parent or
  ;                                    plane, this can be used to stretch a gadget to the edge of the window it
  ;                                    resides on, mod1 specifies additonal space between the left side of the
  ;                                    object and the leftside of the parent, mod2 specifies a minimal width,
  ;                                    you can specify another object (typically a window) or a plane
  ;   #gs_stretchparentright           see above
  ;   #gs_stretchparenttop             see above
  ;   #gs_stretchparentbottom          see above
  ;   #gs_alignparenttop               place the object at the top of the parent, ie. the top of the object will
  ;                                    be directly next to the top of the parent, mod1 specifies additional spacing
  ;   #gs_alignparentbottom            similar to above
  ;   #gs_alignparentleft              similar to above
  ;   #gs_alignparentright             similar to above, modifier2 specifies minimal x coordinate (object will not
  ;                                    be moved to a horizontal position before specified coordinate)
  ;
  ;   special rules                    notes
  ;   -------------------------------  ----------------------------------------------------------------------------
  ;   #gs_parent                       this rule not only affects an exisiting plane object, it will also create
  ;                                    one if the plane object did not yet exist, field usage is as follows:
  ;                                    <plane> = attach rule to a specific plane
  ;                                    <objecttype> = #gs_plane,
  ;                                    <objectnr> = plane affected (it might be created on the fly)
  ;                                    <referencetype> = object type used as parent
  ;                                    <referencenr> = object number used as parent
  ;                                    <modifier1> = plane used as parent (if any)
  ;
  ;
  ; 4. 2D elements.
  ; ---------------
  ;
  ; g_spot keeps track of a set of 2D elements, which can be resized and positioned just like gadgets and windows
  ; using the same ruleset. If 2D elements and gadgets exist on the same plane, they can refer to and
  ; 'interact' with eachother.
  ;
  ; Currently only the element type box can be resized. All elements can be repositioned.
  ;
  ;   element       notes
  ;   ------------  -----------------------------------------------------------------------------------------------
  ;   #gs_shadow    a 2D placeholder for an actual object, for example a gadget, reusing value from above
  ;   #gs_point     single point x y
  ;   #gs_linexy    line x1 y1 to x2 y2
  ;   #gs_box       box at x1 y1 with size width and height
  ;   #gs_circle    circle at x1 y1 with given radius
  ;   #gs_refer     simple element to show relations between objects
  ;   #gs_image     an purebasic image object
  ;
  ; All elements on a plane can be shifted an additional distance by setting gs_planes()\shiftx and y, see 'planes'.
  ; Only elements can be shifted!
  ;
  ;
  ; 4.1 Add an element.
  ; -------------------
  ;
  ; To add an element use:
  ;
  ;   gs_box( <plane>, <x>, <y>, <width>, <height>, <height> [ , <color [ , <mode> ]] )
  ;
  ; Other commands are gs_line() gs_point() gs_circle() and gs_image().
  ;
  ;
  ; 4.2 Add a shadow.
  ; -----------------
  ;
  ; A shadow is a special element. It represents a gadget, and will be drawn using the same coordinates.
  ; A shadow cannot be refered to, or used in any other rule.
  ;
  ;   gs_shadow( <plane> , <objecttype.i> , <objectnr.i> [ , <color.i> [ , <mode.i> [ , <addrule.i> ]]] )
  ;
  ;
  ; 5. Planes.
  ; ----------
  ;
  ; Each 'plane' is an abstract object, more like a frame of reference. Elements and objects 'live' in a plane.
  ; Planes be used to build hierarchical models, for example gadgets on window 1 use a different coordinate set
  ; than the gagets on window 2, or two different CAD drawings can be superimposed on top of eachother.
  ;
  ; Note that gadgets, regardless of the window they exist on, have a unique numbering. The same applies to
  ; elements. You can uniquely identify any object by the combination of type + number, the plane it exists upon
  ; is not relevent for identifying or manipulating a gadget, window or element. However ignoring the 'relative'
  ; position of the plane may lead to unexpected results :-)
  ;
  ; You don't have to create or use planes, it is fine to set everything to plane 0 and ignore this feature if
  ; you are only using the ruleset to modify gadget / window positions, or if you only need a single plane for
  ; 2D elements.
  ;
  ; All elements (but no objects!) can be 'shifted' per plane by an x/y distance by setting gs_planes()\shiftx
  ; and gs_planes()\shifty, and setting gs_enableshift = #true. By default this feature is disabled to speed
  ; things up. Only elements can be shifted. Another way to think of this is consider it a relocation of the
  ; origin point for elements only.
  ;
  ; A plane is an abstract object, and unless rules are set none of its parameters (size, position etc.) will
  ; change.
  ;
EndProcedure

Procedure gs_init()                                                                   ; must be called before using any other g_spot
  ;
  ; *** initialize variables, constants and parameters
  ;
  ; always call this one at the start of your program, call gs_end() at the end
  ;
  #gs_none = -1
  #gs_null = 0
  ;
  Enumeration 1
    ;
    ; each object has four corners x1y1 x2y1 x2y1 x2y2 or four sides x1 x2 y1 y2
    ; these rules allow manipulation of location and sizes, where typical the object is modified depending on coordinates of the reference
    ;
    ; rules
    ;
    ; #gs_none              ; no rule, do nothing
    ; #gs_null              ; pretty much the same thing
    #gs_setx                ; set x coordinate of object, object will move but not resize (x1 object = x1 reference + modifier1)
    #gs_sety                ; set y coordinate
    #gs_setxy               ; set x and y simultaneously
    #gs_setwidth            ; set width (width = width reference + modifier1)
    #gs_setheight           ; set height
    #gs_setsize             ; set with and height
    #gs_use                 ; use x, y, width and height of reference
    #gs_stackup             ; place the object above the reference and match width, modifier1 provides additional spacing upwards
    #gs_stackdown           ; place the object below the reference and match width, modifier1 provides additional spacing downwards
    #gs_stackleft           ; place the object left of the reference and match height, modifier1 provides additional spacing leftwards
    #gs_stackright          ; place the object right of the reference and match height, modifier1 provides additional spacing rightwards
    #gs_alignleft           ; alias for setx
    #gs_alignright          ; align right side of object and reference (x2 object = x2 reference + modifier1)
    #gs_aligntop            ; alias for sety
    #gs_alignbottom         ; align bottom side of object with bottom side of reference (x2 object = x3 reference + modifier)
    #gs_leftof              ; align right side of object with left side of reference, modifier1 provides additional spacing leftwards
    #gs_rightof             ; align left side of object with right side of reference, modifier1 provides additional spacing rightwards
    #gs_above               ; align bottom of object with top of reference, modifier1 provides additional spacing upwards
    #gs_below               ; align top of object with bottom of reference, modifier1 provides additional spacing downwards
    #gs_stretchleft         ; stretch leftside of object to rightside of reference
    #gs_stretchright        ; stretch rightside of object to leftside of reference
    #gs_stretchup           ; stretch topside of object to bottomside of reference
    #gs_stretchdown         ; stretch bottomside of object to topside of reference
    #gs_x1x1                ; alias for setx
    #gs_x1x2                ; alies for rightof
    #gs_x2x1                ; alias for leftof
    #gs_x2x2                ; alias for alignright
    #gs_y1y1                ; alias for sety
    #gs_y1y2                ; alias for below
    #gs_y2y1                ; alias for above
    #gs_y2y2                ; alias for alignbottom
    ; #gs_stretchx1x1         ; not yet implemented
    #gs_stretchx1x2         ; alias for stretchleft
    #gs_stretchx2x1         ; alias for stretchright
    ; #gs_stretchx2x2         ; not implemented yet
    ; #gs_stretchy1y1         ; not implemented yet
    #gs_stretchy1y2         ; alias for stretchup
    #gs_stretchy2y1         ; alias for stretchdown
    #gs_stretchy2y2           ; not yet implemented
    ; #gs_x1w                 ; not yet implemented
    ; #gs_x2w                 ; not yet implemented
    ; #gs_y1h                 ; not yet implemented
    ; #gs_y2h                 ; not yet implemented
    #gs_stretchparentleft    ; stretch the object towards to the left of parent (which could be either a plane, window or gadget)
    #gs_stretchparentright   ; stretch the object towards to the right of parent
    #gs_stretchparenttop     ; strecht the object upwards to the top of parent
    #gs_stretchparentbottom  ; strecht the object downwards to the bottom of parent (
    #gs_alignparenttop       ; place at the top of parent
    #gs_alignparentbottom    ; place at the bottom of parent
    #gs_alignparentleft      ; place at the left side of parent
    #gs_alignparentright     ; place at the right side of parent
    ; #gs_limitx              ; not yet implemented
    ; #gs_limity              ; not yet implemented
    ; #gs_limitwidth          ; not yet implemented
    ; #gs_limitheight         ; not yet implemented
    #gs_parent              ; create a plane using specified parent
    ;
    ; object type / reference types
    ;
    ; #gs_none              ; no reference, often assumed to be the same as null, reusing value from above
    ; #gs_null              ; nothing, no reference, reusing value from above
    #gs_plane               ; a plane (not the flying type :-))
    #gs_window              ; a purebasic window
    #gs_gadget              ; a purebasic gadget
    #gs_image               ; a purebasic image
    #gs_element             ; a 2d element
    #gs_rule                ; used inside refer elements to indicate the rule they represent
    ; #gs_desktop           ; full desktop, not yet implemented
    ;
    ; element-types
    ;
    #gs_shadow              ; a 2D placeholder for an actual object, for example a gadget, reusing value from above
    #gs_point               ; single point x y
    #gs_marker              ; non-scaling square box with the same size as a handle
    #gs_line                ; line x1 y1 to x2 y2
    #gs_box                 ; box at x1 y1 with size width and height
    #gs_circle              ; circle at x1 y1 with given radius
    #gs_refer               ; simple element to show relations between objects
    ;
    ; other constants
    ;
    #gs_panclick
    #gs_pandrag
    #gs_panbox
    #gs_panimage
    #gs_panlive
    ;
  EndEnumeration
  ;
  Structure gs_planes                                                     ; define coordinates and boundary of each plane
    plane.i                                                               ; plane number
    parentplane.i                                                         ; number of parent plane
    x.q                                                                   ; coordinate of left upper corner in relation to parent plane
    y.q                                                                   ; same for y
    width.q                                                               ; width of plane (especially if plane matches an object)
    height.q                                                              ; same for y
    shiftx.q                                                              ; add specfied x to the x-coordinate of all elements on this plane
    shifty.q                                                              ; same for y
    name.s                                                                ; name of the plane
  EndStructure
  ;
  #gs_planesmax = 1000
  ;
  Structure gs_rules                                                      ; structure describing relations between gadgets and / or elements
    plane.i                                                               ; rule belongs to specified plane
    objecttype.i                                                          ; type of object (#gs_adget or #gs_window)
    objectnr.i                                                            ; object number affected
    rule.i                                                                ; rule to apply
    referencetype.i                                                       ; type of object referenced to
    referencenr.i                                                         ; reference object nr
    modifier1.q                                                           ; numeral modifier to reference
    modifier2.q                                                           ; numeral modifier to reference
    inuse.i                                                               ; used to skip failing rules (rule false or object does not exist)
    objectkey.s                                                           ; key to find object back
    referencekey.s                                                        ; key to find reference back
  EndStructure
  ;
  Structure gs_elements                                                   ; structure describing 2d elements not being actual purebasic gadgets or windows
    plane.i                                                               ; rule belongs to specified plane
    shape.i                                                               ; point, line, box or circle
    color.i                                                               ; drawing color (pb frontcolor)
    drawingmode.i                                                         ; drawingmode (pb 2d drawingmodes)
    x.q                                                                   ; x-coordinate
    y.q                                                                   ; y-coordinate
    width.q                                                               ; width
    height.q                                                              ; height
    radius.q                                                              ; radius
    handlex.q                                                             ; handle, typically start point, for circles the circle center
    handley.q                                                             ; handle
    boundx.q                                                              ; bounding box x coordinate
    boundy.q                                                              ; bounding box
    boundwidth.q                                                          ; bounding box
    boundheight.q                                                         ; bounding box
    referencetype.i                                                       ; type of object refered to, see the different reference types above, for example #gs_image
    referencenr.i                                                         ; number of object refered to, for example an image number
  EndStructure
  ;
  Structure gs_objects                                                    ; generated on the fly to keep track of (changed) gadgets and elements
    plane.i                                                               ; plane (element group) the object belongs to
    objecttype.i                                                          ; type of object
    objectnr.i                                                            ; object number affected
    x.q                                                                   ; x-coordinate
    y.q                                                                   ; y-coordinate
    width.q                                                               ; width
    height.q                                                              ; height
    oldx.q                                                                ; old x, to avoid unneccesary redraws
    oldy.q                                                                ; old y
    oldwith.q                                                             ; old width
    oldheight.q                                                           ; old height
  EndStructure
  ;
  Global NewList gs_rules.gs_rules()                                      ; list containing ruleset
  Global NewList gs_elements.gs_elements()                                ; list containing 2d elements
  Global NewList gs_planes.gs_planes()                                    ; list containing planes information
  ;
  Global NewMap gs_objects.gs_objects()                                   ; hashtable containing objects / coordinates
  ;
  ; viewport / viewpane parameters
  ;
  Global gs_viewscale.d = 1.0                             ; default viewscale
  Global gs_viewportx.q = 0                               ; default viewport x
  Global gs_viewporty.q = 0                               ; same for y
  Global gs_viewpanewidth.i = 100                         ; default viewpane width
  Global gs_viewpaneheight.i = 100                        ; same for height
  Global gs_viewpanereferencetype.i = -1                  ; type of object used to size the viewpane
  Global gs_viewpanereferencenr.i = -1                    ; number of object used to size the viewpane
  ;
  ; drawing flags
  ;
  Global gs_enableshift.i = #False                        ; increase speed by disabling feature which allows shifting all elements on specific plane
  Global gs_showhandle.i = #False                         ; show handle (starting point) of objects
  Global gs_handlecolor.i = RGB($FF,$FF,$FF)              ; color of the handle (default white)
  Global gs_handlesize.i = 3                              ; size of the handle
  Global gs_showbound.i = #False                          ; show bounding box for some element types (lines, circles, etc.)
  Global gs_boundcolor.i = RGB($40,$40,$40)               ; color of bounding box
  Global gs_autoshadows.i = #False                        ; auto generate shadow elements during gs_apply() for all objects affected by or referred to by rules
  Global gs_shadowcolor.i = RGB($80,$80,$80)              ; color of shadow elements (a shadow element is a simple square box in light grey)
  Global gs_autorefers.i = #False                         ; auto generate refer objects during gs_apply()
  Global gs_showreferids.i = #False                       ; show gadgets / elements referred to by refers, and show rule numbers in refer elemetns
  Global gs_crookedrefers.i = #True                       ; use a 'crooked' line to display refers (increases visibility)
  Global gs_refercolor.i = RGB($00,$80,$FF)               ; default color for refers unless color has been specified during createion of the refer element
  Global gs_autorefersizecolor.i = RGB($00,$80,$FF)       ; color used to indicate size rules, used by gs_autorefers()
  Global gs_autoreferplacecolor.i = RGB($00,$FF,$80)      ; color used to indicate position rules, used by gs_autorefers()
  Global gs_autoreferspecialcolor.i = RGB($FF,$80,$00)    ; color used to indicate special rules, used by gs_autorefers()
  Global gs_autoscale = #False                            ; enable viewport / viewpane autoscaling during gs_apply()
  Global gs_autopan = #False                              ; enable viewport / viewpane autoscaling during gs_apply()
  ;
  ; other
  ;
  Global gs_internalsviewer_windownr.i = -1               ; used inside the internalsviewer
  Global gs_internalsviewer_gadgetnr.i = -1               ; used inside the internalsviewer
  Global gs_x.q                                           ; used to return a value from some procedures
  Global gs_y.q                                           ; used to return a value from some procedures
  Global gs_windowmousedx.i                               ; used for gs_windowmousex()
  Global gs_windowmousedy.i                               ; used for gs_windowmousey()
  ;
EndProcedure

Procedure gs_end()                                                                    ; clean up api / system stuff that any g_spot procedure may have been using
  ;
  ; *** free up any loose ends used by the g_spot library
  ;
  ; thus far no need :-) as most is handled by purebasic itself... but keep calling this routine
  ; before exiting your program, because i may change things...
  ;
EndProcedure

Macro gs_debug(text)                                                                  ; reports line number + message in the debugger window
  Debug "["+Str(#PB_Compiler_Line)+"] "+text
EndMacro

Procedure gs_setgadgetitem(gadgetnr.i,row.i,text.s)                                   ; overwrite data in a listicongadget()
  Protected n.i, p.i, t.s, column.i, t_old.s, t_new.s
  ;
  ; *** overwrite data in a listicongadget()
  ;
  If row = -1
    row = CountGadgetItems(gadgetnr)
  EndIf
  While row >= CountGadgetItems(gadgetnr)
    AddGadgetItem(gadgetnr,-1,"")
  Wend
  ;
  n = CountString(text,#LF$)+1
  column = 0
  While column < n
    t_new = StringField(text,column+1,#LF$)
    t_old = GetGadgetItemText(gadgetnr,row,column)
    If t_old <> t_new
      SetGadgetItemText(gadgetnr,row,t_new,column)
    EndIf
    column = column+1
  Wend
  ;
  p = 0
  Repeat
    If GetGadgetItemText(gadgetnr,row,column) = ""
      p = p+1
    Else
      SetGadgetItemText(gadgetnr,row.i,"",column)
    EndIf
    column = column+1
  Until p = 3
  ;
EndProcedure

Procedure.s gs_objectkey(objecttype.i,objectnr.i)                                     ; find readable key name for objecttype + objectnumber (for debugging and hash maps)
  Protected key.s
  ;
  ; return readable key for objecttype + objectnumber combination
  ;
  Select objecttype
  Case #gs_none, #gs_null
    key = "none"
  Case #gs_window
    key = "window "+Str(objectnr)
  Case #gs_gadget
    key = "gadget "+Str(objectnr)
  Case #gs_image
    key = "image "+Str(objectnr)
  Case #gs_plane
    key = "plane "+Str(objectnr)
  Case #gs_element
    key = "element "+Str(objectnr)
  Case #gs_rule
    key = "rule "+Str(objectnr)
  Default
    key = "unknown"
    gs_debug("unknown object type")
  EndSelect
  ;
  ProcedureReturn key                                         
EndProcedure

Procedure.s gs_shapename(shapetype)                                                   ; find readable text name for shapetype number
  Protected shape.s
  ;
  ; return readable name for specified element number
  ;
  Select shapetype
  Case #gs_point
    shape = "point"
  Case #gs_marker
    shape = "marker"
  Case #gs_box
    shape = "box"
  Case #gs_line
    shape = "line"
  Case #gs_circle
    shape = "circle"
  Case #gs_image
    shape = "image"
  Case #gs_shadow
    shape = "shadow"
  Case #gs_refer
    shape = "refer"
  Default
    Debug shapetype
    shape = "unknown"
    gs_debug("unknown element shape")
  EndSelect
  ;
  ProcedureReturn shape                                       
EndProcedure

Procedure.s gs_rulename(ruletype)                                                     ; find readable text name for ruletype number
  Protected rule.s
  ;
  Select ruletype
  Case #gs_none
    rule = "none"
  Case #gs_null
    rule = "null"
  Case #gs_setx
    rule = "setx"
  Case #gs_sety
    rule = "sety"
  Case #gs_setxy
    rule = "setxy"
  Case #gs_setwidth
    rule = "setwidth"
  Case #gs_setheight
    rule = "setheight"
  Case #gs_setsize
    rule = "setsize"
  Case #gs_use
    rule = "use"
  Case #gs_stackup
    rule = "stackup"
  Case #gs_stackdown
    rule = "stackdown"
  Case #gs_stackleft
    rule = "stackleft"
  Case #gs_stackright
    rule = "stackright"
  Case #gs_alignleft
    rule = "alignleft"
  Case #gs_alignright
    rule = "alignright"
  Case #gs_aligntop
    rule = "aligntop"
  Case #gs_alignbottom
    rule = "alignbottom"
  Case #gs_leftof
    rule = "leftof"
  Case #gs_rightof
    rule = "rightof"
  Case #gs_above
    rule = "above"
  Case #gs_below
    rule = "below"
  Case #gs_stretchleft
    rule = "stretchleft"
  Case #gs_stretchright
    rule = "stretchright"
  Case #gs_stretchup
    rule = "stretchup"
  Case #gs_stretchdown
    rule = "stretchdown"
  Case #gs_x1x1
    rule = "x1x1"
  Case #gs_x1x2
    rule = "x1x2"
  Case #gs_x2x1
    rule = "x2x1"
  Case #gs_x2x2
    rule = "x2x2"
  Case #gs_y1y1
    rule = "y1y1"
  Case #gs_y1y2
    rule = "y1y2"
  Case #gs_y2y1
    rule = "y2y1"
  Case #gs_y2y2
    rule = "y2y2"
  Case #gs_stretchx1x2
    rule = "stretchx1x2"
  Case #gs_stretchx2x1
    rule = "stretchx2x1"
  Case #gs_stretchy1y2
    rule = "stretchy1y2"
  Case #gs_stretchy2y1
    rule = "stretchy2y1"
  Case #gs_stretchy2y2
    rule = "stretchy2y2"
  Case #gs_stretchparentleft
    rule = "stretchparentleft"
  Case #gs_stretchparentright
    rule = "stretchparentright"
  Case #gs_stretchparenttop
    rule = "stretchparenttop"
  Case #gs_stretchparentbottom
    rule = "stretchparentbottom"
  Case #gs_alignparenttop
    rule = "alignparenttop"
  Case #gs_alignparentbottom
    rule = "alignparentbottom"
  Case #gs_alignparentleft
    rule = "alignparentleft"
  Case #gs_alignparentright
    rule = "alignparentright"
  Case #gs_parent
    rule = "parent"
  Default
    rule = "unknwon"
    gs_debug("unknown ruletype")
  EndSelect
  ;
  ProcedureReturn rule                                        
EndProcedure

Procedure gs_internalsviewer(message.s="")                                            ; display lists and tables, for debugging purposes
  Protected t.s, n.i, line.i, l.i, p.i, x.i, y.i, r.i
  ;
  ; *** show all lists and tables used inside g_spot
  ;
  If IsWindow(gs_internalsviewer_windownr) And gs_internalsviewer_windownr <> 0
  Else
    n = ExamineDesktops()-1
    If n > 0
      gs_internalsviewer_windownr = OpenWindow(#PB_Any,DesktopX(n)+10,DesktopY(n)+10,DesktopWidth(n)-25,DesktopHeight(n)-80, "g_spot structure viewer",#PB_Window_SystemMenu|#PB_Window_SizeGadget)
    Else
      gs_internalsviewer_windownr = OpenWindow(#PB_Any,DesktopWidth(n)-400,40,380,400, "g_spot structure viewer",#PB_Window_SizeGadget|#PB_Window_SystemMenu)
    EndIf
    gs_internalsviewer_gadgetnr = ListIconGadget(#PB_Any,0,0,WindowWidth(gs_internalsviewer_windownr),WindowHeight(gs_internalsviewer_windownr),"",80,#PB_ListIcon_GridLines)
    For n = 1 To 16
      AddGadgetColumn(gs_internalsviewer_gadgetnr,1,"",80)
    Next n
  EndIf
  ;
  ; *** add message
  ;
  r = 0
  If message > ""
    gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
    r = r+1
    gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"*** "+message+" ***")
    r = r+1
  EndIf
  ;
  ;
  ; *** list all planes
  ;
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"gs_planes()" +#LF$+ Str(ListSize(gs_planes())))
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"plane" +#LF$+ "parent" +#LF$+ "x" +#LF$+ "y" +#LF$+ "width" +#LF$+ "height" )
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  ;
  ForEach gs_planes()
    With gs_planes()
      t = Str(\plane) +#LF$+ Str(\parentplane) +#LF$+ Str(\x) +#LF$+ Str(\y) +#LF$+ Str(\width) +#LF$+ Str(\height)
      gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,t)
      r = r+1
    EndWith
  Next
  ;
  ; *** list all rules
  ;
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"gs_rules()" + #LF$+ Str(ListSize(gs_rules())))
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"index" +#LF$+ "plane" +#LF$+ "inuse" +#LF$+ "rule" +#LF$+ "objkey" +#LF$+ "refkey" +#LF$+ "mod1" +#LF$+ "mod2")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  ;
  ForEach gs_rules()
    With gs_rules()
      t = Str(ListIndex(gs_rules())) +#LF$+ Str(\plane) +#LF$+ Str(\inuse) +#LF$
      t = t + gs_rulename(\rule)
      t = t + #LF$ + \objectkey +#LF$+ \referencekey +#LF$+ Str(\modifier1) +#LF$+ Str(\modifier2)
      gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,t)
      r = r+1
    EndWith
  Next
  ;
  ;
  ; *** list all elements
  ;
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"gs_elements()" +#LF$+ Str(ListSize(gs_elements())))
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"index" +#LF$+ "plane" +#LF$+ "key" +#LF$+ "shape" +#LF$+ "x" +#LF$+ "y" +#LF$+ "width" +#LF$+ "height" +#LF$+ "boundx" +#LF$+ "boundy" +#LF$+ "boundwith" +#LF$+ "boundheight" +#LF$+ "handlex" +#LF$+ "handley" +#LF$+ "color" +#LF$+ "drawmode" +#LF$+ "refkey" +#LF$ )
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  ;
  ForEach gs_elements()
    With gs_elements()
      t = Str(ListIndex(gs_elements())) +#LF$+ Str(\plane) +#LF$+ "element "+Str(ListIndex(gs_elements()))+#LF$
      t = t+gs_shapename(\shape)
      t = t+#LF$+ Str(\x) +#LF$+ Str(\y) +#LF$+ Str(\width) +#LF$+ Str(\height) +#LF$+ Str(\boundx) +#LF$+ Str(\boundy) +#LF$+ Str(\boundwidth) +#LF$+ Str(\boundheight) +#LF$+ Str(\handlex) +#LF$+ Str(\handley) +#LF$+ Str(\color) +#LF$+ Str(\drawingmode) +#LF$
      t = t+gs_objectkey(\referencetype,\referencenr)
      gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,t)
      r = r+1
    EndWith
  Next
  ;
  ;
  ; *** list all objects
  ;
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"gs_objects()" + #LF$+ Str(MapSize(gs_objects())))
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"key" +#LF$+ "plane" +#LF$+ "x" +#LF$+ "y" +#LF$+ "width" +#LF$+ "height" +#LF$ )
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  ;
  ForEach gs_objects()
    With gs_objects()
      t = MapKey(gs_objects()) +#LF$+ Str(\plane) +#LF$+ Str(\x) +#LF$+ Str(\y) +#LF$+ Str(\width) +#LF$+ Str(\height)
      gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,t)
      r = r+1
    EndWith
  Next
  ;
  ;
  ; *** show viewport and viewpane information
  ;
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"view")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"")
  r = r+1
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,"view"+#LF$+"scale"+#LF$+"x"+#LF$+"y"+#LF$+"width"+#LF$+"height"+#LF$+"refkey")
  r = r+1
  t = "viewport"+#LF$+Str(gs_viewscale)+#LF$+Str(gs_viewportx)+#LF$+Str(gs_viewporty)+#LF$+Str(gs_viewportx+gs_viewpanewidth/gs_viewscale)
  t = t+#LF$+Str(gs_viewporty+gs_viewpaneheight/gs_viewscale)
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,t)
  r = r+1
  t = "viewpane"+#LF$+#LF$+"0"+#LF$+"0"+#LF$+Str(gs_viewpanewidth)+#LF$+Str(gs_viewpaneheight)+#LF$
  t = t+gs_objectkey(gs_viewpanereferencetype,gs_viewpanereferencenr)
  gs_setgadgetitem(gs_internalsviewer_gadgetnr,r,t)
  r = r+1
EndProcedure

Procedure.i gs_windowmousex(windownr.i)                                               ; return window related mouse coordinate
  Protected wmx.i, dmx.i
  ;
  ; *** return window related mouse coordinate
  ;
  ; the normal windowmousex() returns -1 for any position outside the window, this returns a coordinate relative to the window interior
  ; no matter if the mouse is in- or outside the window
  ;
  wmx = WindowMouseX(windownr)
  dmx = DesktopMouseX()
  If wmx >= 0
    gs_windowmousedx = dmx-wmx
  EndIf
  ProcedureReturn DesktopMouseX()-gs_windowmousedx            
EndProcedure

Procedure.i gs_windowmousey(windownr.i)                                               ; return window related mouse coordinate
  Protected wmy.i, dmy.i
  ;
  ; *** return window related mouse coordinate
  ;
  ; the normal windowmousex() returns -1 for any position outside the window, this returns a coordinate relative to the window interior
  ; no matter if the mouse is in- or outside the window
  ;
  wmy = WindowMouseY(windownr)
  dmy = DesktopMouseY()
  If wmy >= 0
    gs_windowmousedy = dmy-wmy
  EndIf
  ProcedureReturn DesktopMouseY()-gs_windowmousedy            
EndProcedure


; *** planes

Procedure gs_plane(plane.i,parentplane.i,x.q,y.q,width.q,height.q,name.s="",shiftx.q=0,shifty.q=0)  ; create a plane
  Protected found.i
  ;
  ; *** add new plane, or replace / reset plane parameters
  ;
  ; an alternative approach is using the rule #gs_plane using gs_rule() plus gs_apply(), or use gs_parent() where you specify the parent object
  ;
  ResetList(gs_planes())
  With gs_planes()
    While found = #False And NextElement(gs_planes())
      If \plane = plane
        found = #True
      EndIf
    Wend
    If Not found
      LastElement(gs_planes())
      AddElement(gs_planes())
      \plane = plane
    EndIf
    \x = x
    \y = y
    \width = width
    \height = height
    \parentplane = parentplane
    \shiftx = shiftx
    \shifty = shifty
    \name = name
  EndWith
  ;
  SortList(gs_planes(),#PB_Sort_Ascending)
  ;
EndProcedure

Procedure gs_parent(plane.i,parentplane.i,parentobjecttype.i,parentobjectnr.i,name.s="")            ; create a plane
  Protected found.i, retval.i
  ;
  ; *** create a plane by refering to an existing object, this could be for example a gadget, window, etc.
  ;
  retval = #False
  With gs_planes()
    Select parentobjecttype
    Case #gs_plane
      ;
      ; add a plane based on another plane
      ;
      found = #False
      ResetList(gs_planes())
      While NextElement(gs_planes()) And found = #False
        If \plane = parentplane
          found = #True
        EndIf
      Wend
      If found = #True
        gs_plane(plane,parentplane,\x,\y,\width,\height,name)
        retval = #True
      EndIf
      ;
    Case #gs_gadget
      ;
      ; add a plane based on a gadget
      ;
      If IsGadget(parentobjectnr)
        gs_plane(plane,parentplane,GadgetX(parentobjectnr),GadgetY(parentobjectnr),GadgetWidth(parentobjectnr),GadgetHeight(parentobjectnr),name)
        retval = #True
      EndIf
      ;
    Case #gs_window
      ;
      ; add a plane based on a window
      ;
      If IsWindow(parentobjectnr)
        gs_plane(plane,parentplane,WindowX(parentobjectnr),WindowY(parentobjectnr),WindowHeight(parentobjectnr),WindowWidth(parentobjectnr),name)
        retval = #True
      EndIf
      ;
    Case #gs_element
      ;
      ; add a plane based on an element
      ;
      If ListSize(gs_elements()) > parentobjectnr
        SelectElement(gs_elements(),parentobjectnr)
        retval = #True
        gs_plane(plane,parentplane,gs_elements()\x,gs_elements()\y,gs_elements()\width,gs_elements()\height,name)
      EndIf
      ;
    Case #gs_image
      ;
      ; add a plane based on an image
      ;
      If IsImage(parentobjectnr)
        gs_plane(plane,parentplane,0,0,ImageWidth(parentobjectnr),ImageHeight(parentobjectnr),name)
        retval = #True
      EndIf
      ;
    Case #gs_none
      ;
      ; unknown parent type so assume no parent
      ;
      gs_plane(plane,#gs_none,0,0,0,0,name)
      ;
    Case #gs_null
      ;
      ; assume parent is at 0,0 size 0,0
      ;
      gs_plane(plane,parentplane,0,0,0,0,name)
      ;
    Default
      ;
      ; well, nothing here
      ;
      gs_debug("unknown object type")
    EndSelect
  EndWith
  ;
EndProcedure

Procedure.i gs_removeplane(plane.i)                                                   ; remove a plane
  Protected retval.i
  ;
  ; *** remove a specific plane
  ;
  retval = #False
  ForEach gs_planes()
    With gs_planes()
      If \plane = plane
        DeleteElement(gs_planes())
        retval = #True
      EndIf
    EndWith
  Next
  ;
  ProcedureReturn retval                                      
  ;
EndProcedure

Procedure gs_clearallplanes()                                                         ; remove all planes
  ClearList(gs_planes())
EndProcedure

Procedure.i gs_topxy(plane.i,x.q,y.q)                                                 ; traverse planes upwards and find real world coordintes for x,y, plane shifts are ignored
  Protected found.i, l.i
  ;
  ; *** turn x,y into real world x,y
  ;
  ; in:      plane.i         - starting plane
  ;          x.q             - x-coordinate
  ;          y.q             - y-coordinate
  ; retval:  -1              - hit search limit (endless loop, hit 100 levels)
  ;          n               - number of levels traveled up
  ; out:     gs_x.q          - converted x-coordinate
  ;          gs_y.q          - converted y-coordinate
  ;
  ; notes:
  ;
  ; - this function traverses all planes, to find the highest parent
  ; - it will convert the x,y coordinates of the child parent to those of the host parent
  ; - this function limits searchdepth to #gs_planesmax to handle endless loops through an unmanaged list :-)
  ;
  l.i = 0
  With gs_planes()
    found = #False
    ResetList(gs_planes())
    While NextElement(gs_planes()) And found = #False
      If \plane = plane
        x = x+\x
        y = y+\y
        If \parentplane > -1 And \parentplane <> plane And l < #gs_planesmax
          l = l+1
          plane = \parentplane
          ResetList(gs_planes())
          found = #False
        Else
          found = #True
        EndIf
      EndIf
    Wend
  EndWith
  ;
  gs_x = x
  gs_y = y
  If l < #gs_planesmax
    ProcedureReturn l                                           
  Else
    ProcedureReturn -1                                          
  EndIf
  ;
EndProcedure

Procedure gs_shift(plane.i,shiftx.q,shifty.q)                                         ; shift all elements on specified plane
  Protected found.i
  ;
  ;  *** look up and change shifting of elements on the specified plane(s)
  ;
  ; notes:
  ;
  ; - the coordinates of the elements on a plane are not modified, they are only drawn with the specified ofset
  ;
  ResetList(gs_planes())
  found = #False
  While found = #False And NextElement(gs_planes())
    If plane = -1
      gs_planes()\shiftx = shiftx
      gs_planes()\shifty = shifty
    ElseIf gs_planes()\plane = plane
      found = #True
      gs_planes()\shiftx = shiftx
      gs_planes()\shifty = shifty
    EndIf
  Wend
EndProcedure


; *** 2d elements

Procedure.i gs_point(plane.i,x.q,y.q,color.i=-1,mode.i=-1)                            ; add a single point element
  ;
  ; *** create a point element and return its position in the elements list
  ;
  With gs_elements()
    LastElement(gs_elements())
    AddElement(gs_elements())
    \plane = plane
    \shape = #gs_point
    \x = x
    \y = y
    \width = 1
    \height = 1
    \handlex = x
    \handley = y
    \boundx = x
    \boundy = y
    \boundwidth = 1
    \boundheight = 1
    \color = color
    \drawingmode = mode
    \referencetype = #gs_none
    \referencenr = -1
  EndWith
  ;
  ProcedureReturn ListSize(gs_elements())-1                   
EndProcedure

Procedure.i gs_marker(plane.i,x.q,y.q,color.i=-1,mode.i=-1)                           ; add a marker element
  ;
  ; *** create a marker element and return its position in the elements list
  ;
  With gs_elements()
    LastElement(gs_elements())
    AddElement(gs_elements())
    \plane = plane
    \shape = #gs_marker
    \x = x
    \y = y
    \width = 1
    \height = 1
    \handlex = x
    \handley = y
    \boundx = x
    \boundy = y
    \boundwidth = 1
    \boundheight = 1
    \color = color
    \drawingmode = mode
    \referencetype = #gs_none
    \referencenr = -1
  EndWith
  ;
  ProcedureReturn ListSize(gs_elements())-1                   
EndProcedure

Procedure.i gs_linexy(plane.i,x1.q,y1.q,x2.q,y2.q,color.i=-1,mode.i=-1)               ; add a line element
  ;
  ; *** create a line element and return its position in the elements list
  ;
  With gs_elements()
    LastElement(gs_elements())
    AddElement(gs_elements())
    \plane = plane
    \shape = #gs_line
    \color = color
    \drawingmode = mode
    \x = x1
    \y = y1
    \handlex = x1
    \handley = y1
    If x1 < x2
      \boundx = x1
      \boundwidth = x2-x1
      \width = x2-x1
    Else
      \boundx = x2
      \boundwidth = x1-x2
      \width = x2-x1
    EndIf
    If y1 < y2
      \boundy = y1
      \boundheight = y2-y1
      \height = y2-y1
    Else
      \boundy = y2
      \boundheight = y1-y2
      \height = y2-y1
    EndIf
    \referencetype = #gs_none
    \referencenr = -1
  EndWith
  ;
  ProcedureReturn ListSize(gs_elements())-1                   
EndProcedure

Procedure.i gs_box(plane.i,x.q,y.q,width.q,height.q,color.i=-1,mode.i=-1)             ; add a box element
  ;
  ; *** create a box element and return its position in the elements list
  ;
  With gs_elements()
    LastElement(gs_elements())
    AddElement(gs_elements())
    \plane = plane
    \shape = #gs_box
    \color = color
    \drawingmode = mode
    \x = x
    \y = y
    \width = width
    \height = height
    \handlex = x
    \handley = y
    If width < 0
      \boundx = x+width
      \boundwidth = -width
    Else
      \boundx = x
      \boundwidth = width
    EndIf
    If height < 0
      \boundy = y+height
      \boundheight = -height
    Else
      \boundy = y
      \boundheight = height
    EndIf
    \referencetype = #gs_none
    \referencenr = -1
  EndWith
  ;
  ProcedureReturn ListSize(gs_elements())-1                   
EndProcedure

Procedure.i gs_circle(plane.i,x.q,y.q,radius.q,color.i=-1,mode=-1)                    ; add a circle element
  ;
  ; *** create a circle element and return its position in the elements list
  ;
  With gs_elements()
    LastElement(gs_elements())
    AddElement(gs_elements())
    \plane = plane
    \shape = #gs_circle
    \color = color
    \drawingmode = mode
    \x = x
    \y = y
    \width = 2*radius
    \height = 2*radius
    \boundx = \x-radius
    \boundy = \y-radius
    \boundwidth = \width
    \boundheight = \height
    \handlex = \x
    \handley = \y
    \referencetype = #gs_none
    \referencenr = -1
    \radius = radius
  EndWith
  ;
  ProcedureReturn ListSize(gs_elements())-1                   
EndProcedure

Procedure.i gs_image(plane.i,x.q,y.q,width.q,height.q,imagenr.i,mode.i=-1)            ; add an image element
  ;
  ; *** create an image element and return its position in the elements list
  ;
  With gs_elements()
    LastElement(gs_elements())
    AddElement(gs_elements())
    \shape = #gs_image
    \color = 0
    \drawingmode = mode
    \x = x
    \y = y
    \width = width
    \height = height
    \boundx = x
    \boundy = y
    \boundwidth = width
    \boundheight = height
    \handlex = x
    \handley = y
    \referencetype = #gs_image
    \referencenr = imagenr
  EndWith
  ;
  ProcedureReturn ListSize(gs_elements())-1                   
EndProcedure

Procedure.i gs_shadow(plane.i,objecttype.i,objectnr.i,color.i=-1,mode.i=-1,addrule.i=0)                               ; add a placeholder for ie. shadow to a gadget
  Protected retval.i,x.q,y.q,width.q,height.q
  ;
  With gs_elements()
    Select objecttype
    Case #gs_gadget
      If IsGadget(objectnr)
        x = GadgetX(objectnr)
        y = GadgetY(objectnr)
        width = GadgetWidth(objectnr)
        height = GadgetHeight(objectnr)
      EndIf
    Default
      gs_debug("shadows can only refer to gadgets")
      retval = -1
    EndSelect
    ;
    If retval > -1
      LastElement(gs_elements())
      AddElement(gs_elements())
      \plane = plane
      \shape = #gs_shadow
      \color = color
      \drawingmode = mode
      \x = x
      \y = y
      \width = width
      \height = height
      \handlex = x
      \handley = y
      \referencetype = objecttype
      \referencenr = objectnr
      retval = ListSize(gs_elements())-1
      ;
      If addrule
        gs_rule(#gs_element,retval,#gs_shadow,objecttype,objectnr)
      EndIf
      ;
    EndIf
    ;
  EndWith
  ;
  ProcedureReturn retval                                      
EndProcedure

Procedure.i gs_refer(plane.i,objecttype.i,objectnr.i,referencetype.i,referencenr.i,color.i=-1,mode.i=-1,rule.i=-1)    ; marker to indicate relation between two objects using object centers
  Protected retval.i, x1.q, y1.q, x2.q, y2.q, width.q, height.q
  ;
  ; *** create a marker (line, 'refer') showing a relation between two objects, using the object centers
  ;
  ; in:           plane.i              - plane on which the object will be placed
  ;               objecttype.i         - type of object the rule applies on, #gs_gadget, #gs_element, etc.
  ;               objectnr.i           - number of object the rule applies on
  ;               referencetype.i      - object used as reference for the rule
  ;               referencenr.i        - object number used as reference for the rule
  ;               color.i = -1         - optional specific colour
  ;               mode.i = -1          - optional specific drawingmode
  ;               rule.i = -1          - optional rule reference, shown with gs_showreferids = #true
  ; retval:
  ; out:
  ;
  ; notes:
  ;
  ; - a refer is a single line plus arrow shape to indicate the relation between two objects
  ; - refers created using this procedure will point to the center of the selected objects
  ; - with gs_autorefers = #true they are stripped and created inside gs_apply()
  ; - refers are not automatically adjusted by gs_apply()
  ; - this procedure will only accept gadgets and elements
  ; - see also gs_referxy()
  ;
  With gs_elements()
    retval = 0
    ;
    ; first the starting point aka object
    ;
    If objecttype = #gs_gadget
      ;
      ; gadgets
      ;
      If Not IsGadget(objectnr)
        retval = -1                                                           ; gadget does not exist
      Else
        x1 = GadgetX(objectnr)+GadgetWidth(objectnr)/2
        y1 = GadgetY(objectnr)+GadgetHeight(objectnr)/2
      EndIf
      ;
    ElseIf objecttype = #gs_element
      ;
      ; elements
      ;
      If objectnr < ListSize(gs_elements())
        SelectElement(gs_elements(),objectnr)
        If \shape = #gs_shadow Or \shape = #gs_refer                          ; refers cannot link refers and / or shadows
          gs_debug("refers cannot link shadows and / or refers")
          retval = -1
        Else
          x1 = \boundx+\boundwidth/2
          y1 = \boundy+\boundheight/2
        EndIf
      Else
        retval = -1
      EndIf
      ;
    Else
      ;
      ; unknown or unwanted object types
      ;
      gs_debug("unexpected object type")
      retval = -1
      ;
    EndIf
    ;
    ; then the ending point aka reference
    ;
    If referencetype = #gs_gadget
      ;
      ; gadgets
      ;
      If Not IsGadget(referencenr)
        retval = -1                                                           ; gadget does not exist
      Else
        x2 = GadgetX(referencenr)+GadgetWidth(referencenr)/2
        y2 = GadgetY(referencenr)+GadgetHeight(referencenr)/2
      EndIf
      ;
    ElseIf referencetype = #gs_element
      ;
      ; elements
      ;
      If referencenr < ListSize(gs_elements())
        SelectElement(gs_elements(),referencenr)
        If \shape = #gs_shadow Or \shape = #gs_refer                          ; refers cannot link refers and / or shadows
          gs_debug("refers cannot link shadows and / or other refers")
          retval = -1
        Else
          x2 = \boundx+\boundwidth/2
          y2 = \boundy+\boundheight/2
        EndIf
      EndIf
      ;
    Else
      ;
      ; unknown or unwanted object types
      ;
      gs_debug("unexpected object type")
      retval = -1
      ;
    EndIf
    ;
    If retval > -1
      LastElement(gs_elements())
      AddElement(gs_elements())
      \plane = plane
      \shape = #gs_refer
      \color = color
      \drawingmode = mode
      \x = x1
      \y = y1
      \width = x2-x1
      \height = y2-y1
      \referencetype = #gs_rule
      \referencenr = rule
      retval = ListSize(gs_elements())-1
    EndIf
  EndWith
  ;
  ProcedureReturn retval                                      
EndProcedure

Procedure.i gs_referxy(plane.i,x1.q,y1.q,x2.q,y2.q,color.i=-1,mode.i=-1,rule.i=-1)                                    ; marker to indicate relation between two objects using coordinates
  ;
  ; *** create a 'refer' (marker, line) to indicate the relation between two objects using coordinates
  ;
  ; notes:
  ;
  ; - refer elements will be stripped and recreated if the gs_autorefers = #true
  ; - see gs_refer() for additional parameters
  ;
  LastElement(gs_elements())
  AddElement(gs_elements())
  With gs_elements()
    \plane.i = plane
    \shape = #gs_refer
    \color = color
    \drawingmode = mode
    \x = x1
    \y = y1
    \width = x2-x1
    \height = y2-y1
    \referencetype = #gs_rule
    \referencenr = rule
  EndWith
  ProcedureReturn ListSize(gs_elements())-1                   
EndProcedure

Procedure gs_drawelement(elementnr.i,color.i=-1)                                      ; draw specified element in specified color
  Protected x.d, y.d, width.d, height.d, radius.d
  Protected handlex.d, handley.d, boundx.d, boundy.d, boundwidth.d, boundheight.d, found.i, shiftx.q, shifty.q
  ;
  ; *** draw the specified single element on current open output in specified colour
  ;
  ; in:     elementnr.i      - nr of element to draw
  ;         [ color.i = -1 ] - default, use color set in element
  ;                     n    - overrule element color with specified color
  ;
  With gs_elements()
    ;
    ; retrieve element and properties
    ;
    SelectElement(gs_elements(),elementnr)
    ;
    ; retrieve eventual plane shifts
    ;
    shiftx = 0
    shifty = 0
    If gs_enableshift = #True
      ResetList(gs_planes())
      found = #False
      While found = #False And NextElement(gs_planes())
        If gs_planes()\plane = \plane
          found = #True
          shiftx = gs_planes()\shiftx
          shifty = gs_planes()\shifty
        EndIf
      Wend
    EndIf
    ;
    ; retrieve location and size
    ;
    x = (\x-gs_viewportx+shiftx)*gs_viewscale
    y = (\y-gs_viewporty+shifty)*gs_viewscale
    width = \width*gs_viewscale
    height = \height*gs_viewscale
    radius = \radius*gs_viewscale
    ;
    ; draw a bounding box using lines (so there's no need to enforce #pb_2ddrawingmode_outlined)
    ;
    If gs_showbound = #True
      Select \shape
      Case #gs_box, #gs_shadow, #gs_refer, #gs_image
      Default
        boundx = (\boundx-gs_viewportx+shiftx)*gs_viewscale
        boundy = (\boundy-gs_viewporty+shifty)*gs_viewscale
        boundwidth = \boundwidth*gs_viewscale
        boundheight = \boundheight*gs_viewscale
        FrontColor(gs_boundcolor)
        Line(boundx,boundy,boundwidth,1)
        Line(boundx+boundwidth-1,boundy,1,boundheight)
        Line(boundx,boundy,1,boundheight)
        Line(boundx,boundy+boundheight-1,boundwidth,1)
      EndSelect
    EndIf
    ;
    ; handles are small square boxes surrounding the x/y starting point of on object
    ; they are not drawn on refers
    ;
    If gs_showhandle = #True And \shape <> #gs_refer
      handlex = (\handlex-gs_viewportx+shiftx)*gs_viewscale
      handley = (\handley-gs_viewporty+shifty)*gs_viewscale
      FrontColor(gs_handlecolor)
      Line(handlex-gs_handlesize,handley-gs_handlesize,2*gs_handlesize+1,1)
      Line(handlex+gs_handlesize,handley-gs_handlesize,1,2*gs_handlesize+1)
      Line(handlex-gs_handlesize,handley-gs_handlesize,1,2*gs_handlesize+1)
      Line(handlex-gs_handlesize,handley+gs_handlesize,2*gs_handlesize+1,1)
    EndIf
    ;
    ; color on procedure call overrules color given in the element definition
    ; color in the element definition overrules currently active frontcolor, ie.
    ;
    ;   color parameter > element color > active color
    ;
    If color > -1
      FrontColor(color)
    ElseIf \color > -1
      FrontColor(\color)
    EndIf
    If \drawingmode > -1
      DrawingMode(\drawingmode)
    EndIf
    ;
    ; output depends on shape
    ;
    Select \shape
    Case #gs_point
      ;
      ; simple point
      ;
      Line(x,y,1,1)
    Case #gs_marker
      ;
      ; non-scaled square box using handle size
      ;
      Line(x-gs_handlesize,y-gs_handlesize,2*gs_handlesize+1,1)
      Line(x+gs_handlesize,y-gs_handlesize,1,2*gs_handlesize+1)
      Line(x-gs_handlesize,y-gs_handlesize,1,2*gs_handlesize+1)
      Line(x-gs_handlesize,y+gs_handlesize,2*gs_handlesize+1,1)
    Case #gs_line
      ;
      ; simple line
      ;
      LineXY(x,y,x+width,y+height)
    Case #gs_box
      ;
      ; simple box
      ;
      Box(x,y,width,height)
    Case #gs_circle
      ;
      ; simple circle
      ;
      Circle(x,y,radius)
    Case #gs_image
      ;
      ; draw an image (bitmap) to scale
      ;
      If IsImage(\referencenr)
        If \width < 0
          width = ImageWidth(\referencenr)*gs_viewscale
          height = ImageWidth(\referencenr)*gs_viewscale
        EndIf
        DrawImage(ImageID(\referencenr),x,y,width,height)
      EndIf
      ;
    Case #gs_refer
      ;
      ; a 'refer' is a line with arrowhead indicating a relation between two objects
      ;
      If color < 0 And \color < 0
        FrontColor(gs_refercolor)
      EndIf
      If gs_crookedrefers = #True
        LineXY(x,y,x-5,y-8)
        LineXY(x-5,y-8,x+width,y+height)
      Else
        LineXY(x,y,x+width,y+height)
      EndIf
      If gs_showreferids
        DrawText(x+width/2-4,y+height/2-4,Str(\referencenr))
      EndIf
      ;
    Case #gs_shadow
      ;
      ; a shadow element is a simple box representing another object
      ;
      If color < 0
        FrontColor(gs_shadowcolor)
      EndIf
      Box(x,y,width,height)
      If gs_showreferids
        DrawText(x+4,y+4,Str(\referencenr))
      EndIf
      ;
    Default
      gs_debug("unknown element type")
    EndSelect
  EndWith
EndProcedure

Procedure gs_drawallelements(plane.i=-1)                                              ; draw all elements
  Protected n.i
  ;
  ; *** draw all elements on current open output
  ;
  ; (you need to do a startdrawing() before calling this routine)
  ;
  ; step 1: draw all except refers
  ;
  n = 0
  While n < ListSize(gs_elements())
    SelectElement(gs_elements(),n)
    If plane = -1 Or gs_elements()\plane = plane
      Select gs_elements()\shape
      Case #gs_refer
      Default
        gs_drawelement(n)
      EndSelect
    EndIf
    n = n+1
  Wend
  ;
  ; step 2: draw all refers
  ;
  n = 0
  While n < ListSize(gs_elements())
    SelectElement(gs_elements(),n)
    If plane = -1 Or gs_elements()\plane = plane
      Select gs_elements()\shape
      Case #gs_refer
        gs_drawelement(n)
      Default
      EndSelect
    EndIf
    n = n+1
  Wend
  ;
EndProcedure

Procedure gs_clearallelements()                                                       ; clear all elements
  ClearList(gs_elements())
EndProcedure

Procedure gs_referforrule(rule.i)                                                     ; create a refer element for a specified rule
  Protected found.i, x1.q, y1.q, width1.q, height1.q, x2.q, y2.q, width2.q, height2.q, plane1.i, plane2.i
  Protected objecttype1.i, objecttype2.i
  ;
  ; *** create a refer element for a specified rule
  ;
  ; in:      rule.i    - rule number to generate the refer for
  ;
  ; notes:
  ;
  ; - gs_apply() must have been called before to fill in the gs_objects() table
  ; - rule and objects must exist, only limited error handling in here
  ; - the generated refer object has the same plane as the specified rule
  ;
  If SelectElement(gs_rules(),rule)
    With gs_objects()
      If gs_rules()\inuse
        ;
        ; find 'real world' coordinates for object and reference
        ;
        found = #False
        If FindMapElement(gs_objects(),gs_rules()\referencekey)
          plane2 = \plane
          width2 = \width
          height2 = \height
          objecttype2 = gs_rules()\referencetype
          gs_topxy(\plane,\x,\y)
          x2 = gs_x
          y2 = gs_y
          found = #True
        EndIf
        If found = #True And FindMapElement(gs_objects(),gs_rules()\objectkey)
          plane1 = \plane
          width1 = \width
          height1 = \height
          objecttype1 = gs_rules()\referencetype
          gs_topxy(\plane,\x,\y)
          x1 = gs_x
          y1 = gs_y
          found = #True
        EndIf
        If found = #True
          ;
          x2 = \x+x2-x1
          y2 = \y+y2-y1
          x1 = \x
          y1 = \y
          ;
          Select gs_rules()\rule
          Case #gs_none, #gs_null
          Case #gs_setwidth
            gs_referxy(plane1,x1+width1/2,y1,x2+width2/2,y2,gs_autorefersizecolor,-1,rule)
          Case #gs_setheight
            gs_referxy(plane1,x1,y1+height1/2,x2,y2+height2/2,gs_autorefersizecolor,-1,rule)
          Case #gs_setsize
            gs_referxy(plane1,x1+width1/2,y1+height1/2,x2+width2/2,y2+height2/2,gs_autorefersizecolor,-1,rule)
          Case #gs_setx
            gs_referxy(plane1,x1,y1,x2,y2+height2/2,gs_autoreferplacecolor,-1,rule)
          Case #gs_sety, #gs_aligntop
            gs_referxy(plane1,x1,y1,x2+width2/2,y2,gs_autoreferplacecolor,-1,rule)
          Case #gs_setxy
            gs_referxy(plane1,x1,y1,x2,y2,gs_autoreferplacecolor,-1,rule)
          Case #gs_use
            gs_referxy(plane1,x1+width1/2,y1+height1/2,x2+width2/2,y2+height2/2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_alignleft
            gs_referxy(plane1,x1,y1,x2,y2+height2/2,gs_autoreferplacecolor,-1,rule)
          Case #gs_alignright
            gs_referxy(plane1,x1,y1,x2+width2,y2+height2/2,gs_autoreferplacecolor,-1,rule)
          Case #gs_alignbottom
            gs_referxy(plane1,x1,y1,x2+width2/2,y2+height2,gs_autoreferplacecolor,-1,rule)
          Case #gs_stackup
            gs_referxy(plane1,x1,y1+height1,x2,y2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stackdown
            gs_referxy(plane1,x1,y1,x2,y2+height2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stackleft
            gs_referxy(plane1,x1+width1,y1,x2,y2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stackright
            gs_referxy(plane1,x1,y1,x2+width2,y2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_above
            gs_referxy(plane1,x1,y1,x2+width2/2,y2,gs_autoreferplacecolor,-1,rule)
          Case #gs_below
            gs_referxy(plane1,x1,y1,x2+width2,y2+height2,gs_autoreferplacecolor,-1,rule)
          Case #gs_rightof
            gs_referxy(plane1,x1,y1,x2+width2,y2+height2/2,gs_autoreferplacecolor,-1,rule)
          Case #gs_leftof
            gs_referxy(plane1,x1,y1,x2,y2+height2/2,gs_autoreferplacecolor,-1,rule)
          Case #gs_stretchup
            gs_referxy(plane1,x1+width1/2,y1,x2+width2/2,y2+height2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stretchdown
            gs_referxy(plane1,x1+width1/2,y1+height1,x2+width2/2,y2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stretchleft
            gs_referxy(plane1,x1,y1+height1/2,x2+width2,y2+height2/2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stretchright
            gs_referxy(plane1,x1+width1,y1+height1/2,x2,y2+height2/2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_setwidth
            gs_referxy(plane1,x1+width1/2,y1,x1+width1/2,0,gs_autorefersizecolor,-1,rule)
          Case #gs_setheight
            gs_referxy(plane1,x1,y1+height1/2,0,y1+height1/2,gs_autorefersizecolor,-1,rule)
          Case #gs_setsize
            gs_referxy(plane1,x1,y1+height1/2,0,height2/2,gs_autorefersizecolor,-1,rule)
          Case #gs_alignparenttop
            gs_referxy(plane1,x1+width1/2,0,x1+width1/2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stretchparentbottom
            gs_referxy(plane1,x1+width1/2,y1+height1,x1+width1/2,height2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stretchparentleft
            gs_referxy(plane1,x1,y1+height1/2,0,y1+height1/2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_stretchparentright
            gs_referxy(plane1,x1+width1,y1+height1/2,width2,y1+height1/2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_alignparenttop
            gs_referxy(plane1,x1+width1/2,0,x1+width1/2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_alignparentbottom
            gs_referxy(plane1,x1+width1/2,y1+height1,x1+width1/2,height2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_alignparentleft
            gs_referxy(plane1,x1,y1+height1/2,0,y1+height1/2,gs_autoreferspecialcolor,-1,rule)
          Case #gs_alignparentright
            gs_referxy(plane1,x1+width1,y1+height1/2,width2,y1+height1/2,gs_autoreferspecialcolor,-1,rule)
          Default
            gs_debug("missing refer definition")
          EndSelect
        EndIf
      EndIf
    EndWith
  EndIf
  ;
EndProcedure

Procedure gs_autorefers(plane.i=-1)                                                   ; create refer elements for all objects on specified plane
  ;
  With gs_elements()
    ForEach gs_elements()
      If \shape = #gs_refer And \plane = plane
        DeleteElement(gs_elements())
      EndIf
    Next
  EndWith
  ;
  ForEach gs_rules()
    If plane = -1 Or gs_rules()\plane = plane
      gs_referforrule(ListIndex(gs_rules()))
    EndIf
  Next
  ;
EndProcedure


; *** viewing and zooming

Procedure gs_setviewpane(width.q,height.q,objecttype.i=-1,objectnr.i=-1)              ; set viewpane size
  ;
  ; *** set physical viewpane size, either by size or by reference
  ;
  ; in:             width.q             - width of viewpane
  ;                 height.q            - height of viewpane
  ;                 objecttype.i        - objecttype #gs_window #gs_gadget or #gs_image
  ;                 objectnr.i          - object refered to
  ; retval:                             - none
  ; out:            gs_viewpanewidth.i  - width of viewpane, in pixels
  ;                 gs_viewpaneheight.i - height of viewpane
  ;                 gs_viewpanetype.i   - object type used to set viewpane size, -1 if not in use or not existing
  ;                 gs_viewpanenr.i     - object number used to set viewpane size
  ; notes:
  ;
  ; - the 'viewport' is a rectangular area whose contents are to be drawn on screen
  ; - the 'viewpane' is an area on the real screen / image etc. onto wich everything in the viewport is drawn
  ; - so x1,y1 (left upper corner) of the viewport matches x1,y1 of the viewpane
  ; - the width and height of the viewport is defined by the width and height of the viewpane
  ;
  gs_viewpanewidth = width
  gs_viewpaneheight = height
  gs_viewpanereferencetype = objecttype
  gs_viewpanereferencenr = objectnr
  ;
  Select objecttype
  Case #gs_window
    If IsWindow(objectnr)
      gs_viewpanewidth = WindowWidth(objectnr)
      gs_viewpaneheight = WindowHeight(objectnr)
    Else
      gs_viewpanewidth = 0
      gs_viewpanereferencetype = #gs_none
    EndIf
  Case #gs_gadget
    If IsGadget(objectnr)
      gs_viewpanewidth = GadgetWidth(objectnr)
      gs_viewpaneheight = GadgetHeight(objectnr)
    Else
      gs_viewpanewidth = 0
      gs_viewpanereferencetype = #gs_none
    EndIf
  Case #gs_image
    If IsImage(objectnr)
      gs_viewpanewidth = ImageWidth(objectnr)
      gs_viewpaneheight = ImageHeight(objectnr)
    Else
      gs_viewpanewidth = 0
      gs_viewpanereferencetype = #gs_none
    EndIf
  Default
    gs_debug("not yet implemented")
  EndSelect
  ;
EndProcedure

Procedure.q gs_worldx(x.i)                                                            ; convert window to world coordinate
  ;
  ; *** convert a window coordinate to a world coordinate
  ;
  ; notes:
  ;
  ; - this assumes the null point of the viewpane lies exactly on the top left corner of the window interior
  ; - if that is not the case, the interiorx and y have to be adjusted, ie. interiorx = x - gadgetx(n)
  ;
  ;
  ; convert a viewport coordinate to a viewpane coordinate
  ;
  ;   x = ( worldx - gs_viewportx ) * gs_viewscale
  ;   y = ( worldy - gs_viewporty ) * gs_viewscale
  ;
  ; convert a viewpane coordinate to a viewport coordinate
  ;
  ;   worldx = x / gs_viewscale + gs_viewportx
  ;   worldy = y / gs_viewscale + gs_viewporty
  ;
  ;
  ProcedureReturn (x/gs_viewscale+gs_viewportx)               
EndProcedure

Procedure.q gs_worldy(y.i)                                                            ; convert window to world coordinate
  ;
  ProcedureReturn (y/gs_viewscale+gs_viewporty)               
EndProcedure

Procedure gs_zoom(factor.d)                                                           ; zoom in / out by percentage
  Protected xc.q, yc.q
  ;
  xc = gs_viewportx+gs_viewpanewidth/gs_viewscale/2
  yc = gs_viewporty+gs_viewpaneheight/gs_viewscale/2
  gs_viewscale = gs_viewscale*factor
  gs_viewportx = xc-gs_viewpanewidth/gs_viewscale/2
  gs_viewporty = yc-gs_viewpaneheight/gs_viewscale/2
  ;
EndProcedure

Procedure gs_zoomarea(x.q,y.q,width.q,height.q)                                       ; zoom in on the selected area
  Protected h_scale.d, v_scale.d
  ;
  ; zoom in on the specified area
  ; make sure the selected area fits in the viewpane and viewport
  ;
  ;
  h_scale = gs_viewpanewidth/width
  v_scale = gs_viewpaneheight/height
  If v_scale < h_scale
    gs_viewscale = v_scale
  Else
    gs_viewscale = h_scale
  EndIf
  ;
  gs_viewportx = x+width/2-gs_viewpanewidth/gs_viewscale/2
  gs_viewporty = y+height/2-gs_viewpaneheight/gs_viewscale/2
  ;
EndProcedure

Procedure gs_zoomall(plane.i=-1)                                                      ; move / scale the viewport so all elements on selected plane become visible
  Protected x1.q, y1.q, x2.q, y2.q, h_scale.d, v_scale.d
  Protected shiftx.q, shifty.q, found.i
  ;
  ; *** zoom out to show all elements on selected plane
  ;
  ; step 1: get boundary sizes of all elements excluding refers and planes
  ;
  x1 = Pow(2,61)
  x2 = -Pow(2,61)
  y1 = Pow(2,61)
  y2 = -Pow(2,61)
  ;
  ForEach gs_elements()
    With gs_elements()
      If plane = -1 Or \plane = plane
        If \shape <> #gs_refer And \shape <> #gs_plane
          ;
          ; retrieve eventual planewise shifts for elements
          ;
          shiftx = 0
          shifty = 0
          If gs_enableshift = #True
            ResetList(gs_planes())
            found = #False
            While found = #False And NextElement(gs_planes())
              If gs_planes()\plane = \plane
                found = #True
                shiftx = gs_planes()\shiftx
                shifty = gs_planes()\shifty
              EndIf
            Wend
          EndIf
          ;
          ; check all object boundaries
          ;
          If \boundx+shiftx < x1
            x1 = \boundx+shiftx
          EndIf
          If \boundx+\boundwidth+shiftx > x2
            x2 = \boundx+\boundwidth+shiftx
          EndIf
          If \boundy+shifty < y1
            y1 = \boundy+shifty
          EndIf
          If \boundy+\boundheight+shifty > y2
            y2 = \boundy+\boundheight+shifty
          EndIf
          ;
        EndIf
      EndIf
    EndWith
  Next
  ;
  ; step 2: calculate optimal scale and left upper corner coordinate
  ;
  h_scale = gs_viewpanewidth/(x2*1.1-x1/1.1)
  v_scale = gs_viewpaneheight/(y2*1.1-y1/1.1)
  ;
  If v_scale < h_scale
    h_scale = v_scale
  EndIf
  gs_viewscale = h_scale
  gs_viewportx = x1+(x2-x1)/2-gs_viewpanewidth/gs_viewscale/2
  gs_viewporty = y1+(y2-y1)/2-gs_viewpaneheight/gs_viewscale/2
  ;
EndProcedure

Procedure gs_zoommouse(windownr.i,gadgetnr.i,imagenr.i)                               ; mark area with the mouse and zoom in
  Protected startx.i, starty.i, oldx.i, oldy.i, event.i, copynr.i, x.i, y.i
  ;
  ; in:      windownr.i = n         - window where the output(gadget) is placed
  ;          gadgetnr.i = n         - imagegadget containing the image
  ;          imagenr.i  = n         - image used to draw upon
  ; retval:  none
  ; out:     none
  ;
  ;
  ; click on a spot and select the area
  ; a box will be drawn to represent the selection
  ;
  ; grab all mouse event messages
  ;
  SetCapture_(WindowID(windownr))
  ;
  ; store start position
  ;
  startx = gs_windowmousex(windownr)-GadgetX(gadgetnr)
  starty = gs_windowmousey(windownr)-GadgetY(gadgetnr)
  ;
  ; create a backup
  ;
  copynr = CopyImage(imagenr,#PB_Any)
  ;
  ; repeat until mouse is released
  ;
  Repeat
    event = WaitWindowEvent()
    x = gs_windowmousex(windownr)-GadgetX(gadgetnr)
    y = gs_windowmousey(windownr)-GadgetY(gadgetnr)
    If x <> oldx Or y <> oldy
      StartDrawing(ImageOutput(imagenr))
        DrawingMode(#PB_2DDrawing_Default)
        DrawImage(ImageID(copynr),0,0)
        DrawingMode(#PB_2DDrawing_Default|#PB_2DDrawing_Outlined)
        FrontColor(RGB($E0,$E0,$E0))
        Box(startx,starty,x-startx,y-starty)
      StopDrawing()
      SetGadgetState(gadgetnr,ImageID(imagenr))
      oldx = x
      oldy = y
    EndIf
  Until event = 514
  ;
  If x < startx
    Swap startx, x
  EndIf
  If y < starty
    Swap starty, y
  EndIf
  gs_zoomarea(gs_worldx(startx),gs_worldy(starty),(x-startx)/gs_viewscale,(y-starty)/gs_viewscale)
  FreeImage(copynr)
  ;
  ReleaseCapture_()
  ;
EndProcedure

Procedure gs_pan(xd.d,yd.d)                                                           ; pan view by specified distance
  ;
  ; *** pan view horizontal or vertical
  ;
  ; in:     xd.d   - horizontal displacement
  ;         yd.d   - vertical displacement
  ;
  gs_viewportx = gs_viewportx+xd*gs_viewpanewidth/gs_viewscale
  gs_viewporty = gs_viewporty+yd*gs_viewpanewidth/gs_viewscale
  ;
EndProcedure

Procedure gs_panto(x.q,y.q)                                                           ; set center to specific coordinates
  ;
  ; *** set center of drawing to specified locations
  ;
  gs_viewportx = x-gs_viewpanewidth/gs_viewscale/2
  gs_viewporty = y-gs_viewpaneheight/gs_viewscale/2
  ;
EndProcedure

Procedure gs_panmouse(mode.i,windownr.i,gadgetnr.i,imagenr.i,plane.i=-1)              ; pan by mousedrag
  Protected startx.i, starty,i, x.i, y.i
  Protected event.i, event_type.i, event_gadget.i
  Protected oldx.i, oldy.i, copynr.i
  ;
  ; in:      mode.i = #gs_panclick  - click on a spot, view will be centered around that spot
  ;                   #gs_pandrag   - drag a distance, no optical feedback
  ;                   #gs_panbox    - drag a distance, a box is drawn as optical feedback
  ;                   #gs_panimage  - drag and draw using a copy of the image
  ;                   #gs_panlive   - drag and draw whilst redrawing active data (slow)
  ;          windownr.i = n         - window where the output(gadget) is placed
  ;          gadgetnr.i = n         - imagegadget containing the image
  ;          imagenr.i  = n         - image used to draw upon
  ;          plane.i    = n         - plane to draw, only used for #gs_live
  ; retval:  none
  ; out:     none
  ;
  ; notes:
  ;
  ; - this routine is called after a leftmousebutton down has been detected
  ; - valid window number, an active imagegadget and an associated bitmap are required
  ; - no error checking!
  ;
  ; store starting position
  ;
  startx = WindowMouseX(windownr)
  starty = WindowMouseY(windownr)
  ;
  ; capture all mouseevents
  ;
  SetCapture_(WindowID(windownr))
  ;
  ; the actual actions depend on the pan mode used
  ;
  Select mode
  Case #gs_panclick
    ;
    ; click on a spot and pan to it
    ;
    Repeat
      event = WaitWindowEvent()
    Until event = 514
    x = gs_windowmousex(windownr)-GadgetX(gadgetnr)
    y = gs_windowmousey(windownr)-GadgetY(gadgetnr)
    x = gs_worldx(x)
    y = gs_worldy(y)
    gs_panto(x,y)
    ;
  Case #gs_pandrag
    ;
    ; click on a spot and drag
    ; no optical feedback
    ;
    Repeat
      event = WaitWindowEvent()
    Until event = 514
    x = gs_windowmousex(windownr)
    y = gs_windowmousey(windownr)
    gs_viewportx = (startx-x)/gs_viewscale+gs_viewportx
    gs_viewporty = (starty-y)/gs_viewscale+gs_viewporty
    ;
  Case #gs_panbox
    ;
    ; click on a spot and drag
    ; a box will be drawn to represent movement
    ; action is opposite of other pan movements!
    ;
    copynr = CopyImage(imagenr,#PB_Any)
    Repeat
      event = WaitWindowEvent()
      x = gs_windowmousex(windownr)
      y = gs_windowmousey(windownr)
      If x <> oldx Or y <> oldy
        StartDrawing(ImageOutput(imagenr))
          DrawingMode(#PB_2DDrawing_Default)
          DrawImage(ImageID(copynr),0,0)
          DrawingMode(#PB_2DDrawing_Default|#PB_2DDrawing_Outlined)
          FrontColor(RGB($E0,$E0,$E0))
          Box(x-startx+1,y-starty+1,gs_viewpanewidth-2,gs_viewpaneheight-2)
        StopDrawing()
        SetGadgetState(gadgetnr,ImageID(imagenr))
        oldx = x
        oldy = y
      EndIf
    Until event = 514
    gs_viewportx = (x-startx)/gs_viewscale+gs_viewportx
    gs_viewporty = (y-starty)/gs_viewscale+gs_viewporty
    FreeImage(copynr)
    ;
  Case #gs_panimage
    ;
    ; click on a spot and drag a graphical representation
    ; an image will be drawn to represent movement
    ;
    copynr = CopyImage(imagenr,#PB_Any)
    oldx = startx
    oldy = starty
    Repeat
      event = WaitWindowEvent()
      x = gs_windowmousex(windownr)
      y = gs_windowmousey(windownr)
      If x <> oldx Or y <> oldy
        StartDrawing(ImageOutput(imagenr))
          DrawingMode(#PB_2DDrawing_Default)
          FrontColor(0)
          Box(0,0,gs_viewpanewidth,gs_viewpaneheight,0)
          DrawImage(ImageID(copynr),x-startx,y-starty)
        StopDrawing()
        SetGadgetState(gadgetnr,ImageID(imagenr))
        oldx = x
        oldy = y
      EndIf
    Until event = 514
    gs_viewportx = (startx-x)/gs_viewscale+gs_viewportx
    gs_viewporty = (starty-y)/gs_viewscale+gs_viewporty
    FreeImage(copynr)
    ;
  Case #gs_panlive
    ;
    ; click on a spot and drag, the image will be real time updated
    ;
    oldx = startx
    oldy = starty
    Repeat
      event = WaitWindowEvent()
      x = gs_windowmousex(windownr)
      y = gs_windowmousey(windownr)
      If x <> oldx Or y <> oldy
        gs_viewportx = (oldx-x)/gs_viewscale+gs_viewportx
        gs_viewporty = (oldy-y)/gs_viewscale+gs_viewporty
        StartDrawing(ImageOutput(imagenr))
          DrawingMode(#PB_2DDrawing_Default)
          FrontColor(0)
          Box(0,0,gs_viewpanewidth,gs_viewpaneheight)
          gs_drawallelements(plane)
        StopDrawing()
        SetGadgetState(gadgetnr,ImageID(imagenr))
        oldx = x
        oldy = y
      EndIf
    Until event = 514
    ;
  EndSelect
  ;
  ReleaseCapture_()
  ;
EndProcedure


; *** rules and resizing

Procedure gs_rule(plane.i,rule.i,objecttype.i,objectnr.i,referencetype.i=-1,referencenr.i=-1,modifier1.q=0,modifier2.q=0)    ; add rule to rulelist
  ;
  ; in:      plane.i                    - plane object belongs to
  ;          objecttype.i               - typ of object affected (#gs_gadget, #gs_window, #gs_element)
  ;          objectnr.i                 - purebasic object number (window number, gadget number etc.)
  ;          rule.i                     - rule applied
  ;          [ referencetype.i ]        - type of object refered to, default -1 no object
  ;          [ referencenr.i ]          - object referred to, default -1 no object
  ;          [ modifier1.i ]            - value added to a coordinate retrieved from the reference object, default 0
  ;          [ modifier2.i ]            - value added to a coordinate retrieved from the reference object, default 0
  ;
  ; note: no error checking!
  ;
  LastElement(gs_rules())
  AddElement(gs_rules())
  With gs_rules()
    \plane = plane
    \rule = rule
    \objecttype = objecttype
    \objectnr = objectnr
    If referencetype = -1 Or referencetype = #gs_none Or referencetype = #gs_null Or referencenr = -1
      \referencetype = #gs_none
      \referencenr = -1
    Else
      \referencetype = referencetype
      \referencenr = referencenr
    EndIf
    \modifier1 = modifier1
    \modifier2 = modifier2
  EndWith
  ;
EndProcedure

Procedure.i gs_removerule(objecttype.i,objectnr.i)                                    ; clear all rules for specific object
  ForEach gs_rules()
    If gs_rules()\objectnr = objectnr And gs_rules()\objecttype = objecttype
      DeleteElement(gs_rules())
    EndIf
  Next
EndProcedure

Procedure.i gs_clearallrules()                                                        ; clear all rules
  ;
  ; *** clear ruleset
  ;
  ClearList(gs_rules())
  ;
EndProcedure

Procedure.i gs_apply(plane.i=-1)                                                      ; apply all valid rules to all known and existing objects
  Protected objectnr.i, referencenr.i, key.s, x.q, y.q, width.q, height.q
  Protected x1.q, y1.q, x2.q, y2.q, changes.i, objecttype.i, found.i
  Protected modifier1.q, modifier2.q
  Protected oldviewscale.d, oldviewportxc.q, oldviewportyc.q, oldviewpanewidth.i, oldviewpaneheight.i
  ;
  ; *** apply all valid rules to all known and existing objects
  ;
  ; in:      [ plane = n ] - only process rules for specified plane
  ;                    -1  - process all planes
  ; retval:  #true         - one or more objects were affected by the ruleset
  ;          #false        - no objects were affected
  ; out:
  ;
  ; notes:
  ;
  ; - resizes and / or or moves images, gadgets, windows, box elements, move non-box elements
  ; - can use images, gadgets, windows, and non shadow elements as reference
  ; - if gs_autoshadows = #true all old shadow elements will be removed, and new ones created for all gadgets in use and existing
  ;
  ;
  ; this routine goes throught he following steps:
  ;
  ; - step 1: clear objects list and add null and none objects
  ; - step 2: remove shadow and refer elements from gs_elements()
  ; - step 3: create objects and / or elements by rules
  ; - step 4: create objects coordinate list gs_objects()
  ; - step 5: store old coordinates to detect need for move or resize, in gs_objects()
  ; - step 6: process all rules, using gs_rules()
  ; - step 7: apply changes
  ; - step 8: create shadow elements
  ;
  ;
  changes = #False                                                        ; default return value
  If ListSize(gs_rules()) > 0
    ;
    ; step 0: preper for autopan and autoscale
    ;
    oldviewscale = gs_viewscale                                           ; current view scale
    oldviewportxc = gs_viewportx+gs_viewpanewidth/gs_viewscale/2          ; these are the center coordinates of the current viewport
    oldviewportyc = gs_viewporty+gs_viewpaneheight/gs_viewscale/2
    oldviewpaneheight = gs_viewpaneheight                                 ; size of current viewpane
    oldviewpanewidth = gs_viewpanewidth
    ;
    ; step 1: clear the objects list and add two 'null' objects
    ;
    ClearMap(gs_objects())
    With gs_objects()
      key = "none"
      AddMapElement(gs_objects(),key)
      \plane = -1
      \x = 0
      \y = 0
      \width = 0
      \height = 0
      \objecttype = #gs_none
      \objectnr = -1
      key = "null"
      AddMapElement(gs_objects(),key)
      \plane = -1
      \x = 0
      \y = 0
      \width = 0
      \height = 0
      \objecttype = #gs_null
      \objectnr = -1
    EndWith
    ;
    ; step 2: remove shadow elements and refer elements
    ;
    If gs_autoshadows
      ForEach gs_elements()
        With gs_elements()
          Select \shape
          Case #gs_shadow
            DeleteElement(gs_elements())
          EndSelect
        EndWith
      Next
    EndIf
    If gs_autorefers
      ForEach gs_elements()
        With gs_elements()
          Select \shape
          Case #gs_refer
            DeleteElement(gs_elements())
          EndSelect
        EndWith
      Next
    EndIf
    ;
    ; step 3: create objects by rules
    ;
    ForEach gs_rules()
      With gs_rules()
        Select \rule
        Case #gs_parent
          ;
          ; create plane number <\objectnumber> using location and size of referred object and use <\modifier1> as parent
          ;
          gs_parent(\objectnr,\modifier1,\referencetype,\referencenr)
          ;
        EndSelect
      EndWith
    Next
    ;
    ; step 4: create objects coordinate list
    ;
    ForEach gs_rules()
      With gs_objects()                                                     ; unless otherwise specified all references are to gs_objects()
        If plane = -1 Or gs_rules()\plane = plane                           ; only process selected planes
          ;
          ; find information on objects for each rule
          ;
          objectnr = gs_rules()\objectnr
          Select gs_rules()\objecttype
          Case #gs_gadget                                                   ; it is a a gadget
            If IsGadget(objectnr)                                           ; does the gadget exist?
              key = gs_objectkey(#gs_gadget,objectnr)                         ; yes, so create a key
              If FindMapElement(gs_objects(),key)                           ; do we already know about it?
                gs_rules()\objectkey = key                                  ; yes, so store the key
                gs_rules()\inuse = #True                                    ; flag the rule as valid
              Else                                                          ; we didn't find this gadget, let's add it
                AddMapElement(gs_objects(),key)                             ; gadget exists, add gadget to gs_objects() table
                gs_rules()\objectkey = key                                  ; store the key to find back the object
                gs_rules()\inuse = #True
                \plane = gs_rules()\plane
                \objecttype = #gs_gadget
                \objectnr = objectnr
                \x = GadgetX(objectnr)
                \y = GadgetY(objectnr)
                \width = GadgetWidth(objectnr)
                \height = GadgetHeight(objectnr)
              EndIf
            Else
              gs_rules()\inuse = #False
            EndIf
          Case #gs_window                                                   ; it is a window
            If IsWindow(objectnr)                                           ; does the window exist?
              key = gs_objectkey(#gs_window,objectnr)                         ; yes, so create a key
              If FindMapElement(gs_objects(),key)                           ; do we already know about it?
                gs_rules()\objectkey = key                                  ; store the key
                gs_rules()\inuse = #True                                    ; flag the rule as valid
              Else                                                          ; we didn't find this window, let's add it
                AddMapElement(gs_objects(),key)                             ; window exists, add window to gs_objects() table
                gs_rules()\objectkey = key                                  ; store the key to find back the object
                gs_rules()\inuse = #True
                \plane = gs_rules()\plane
                \objecttype = #gs_window
                \objectnr = objectnr
                \x = WindowX(objectnr)
                \y = WindowY(objectnr)
                \width = WindowWidth(objectnr)
                \height = WindowHeight(objectnr)
              EndIf
            Else
              gs_rules()\inuse = #False
            EndIf
          Case #gs_element
            If ListSize(gs_elements()) > objectnr                           ; avoid errors, is the list big enough?
              key = gs_objectkey(#gs_element,objectnr)                        ; yes, so create a key
              If FindMapElement(gs_objects(),key)                           ; do we already know about it?
                gs_rules()\objectkey = key                                  ; store the key
                gs_rules()\inuse = #True                                    ; flag the rule as valid
              Else                                                          ; we didn't find this element, let's add it
                AddMapElement(gs_objects(),key)                             ; element exists, add element to gs_objects() table
                gs_rules()\objectkey = key                                  ; store the key to find back the object
                gs_rules()\inuse = #True
                \plane = gs_rules()\plane
                \objecttype = #gs_element
                \objectnr = objectnr
                SelectElement(gs_elements(),objectnr)
                \x = gs_elements()\boundx
                \y = gs_elements()\boundy
                \width = gs_elements()\boundwidth
                \height = gs_elements()\boundheight
              EndIf
            Else
              gs_rules()\inuse = #False
            EndIf
          Case #gs_image
            If IsImage(objectnr)                                            ; avoid errors, does the image exist?
              key = gs_objectkey(#gs_image,objectnr)                          ; yes, so create a key
              If FindMapElement(gs_objects(),key)                           ; do we already know about it?
                gs_rules()\objectkey = key                                  ; store the key
                gs_rules()\inuse = #True                                    ; flag the rule as valid
              Else                                                          ; we didn't find this image, let's add it
                AddMapElement(gs_objects(),key)                             ; image exists, add image to gs_objects() table
                gs_rules()\objectkey = key                                  ; store the key to find back the object
                gs_rules()\inuse = #True
                \plane = gs_rules()\plane
                \objecttype = #gs_image
                \objectnr = objectnr
                \x = 0
                \y = 0
                \width = ImageWidth(objectnr)
                \height = ImageHeight(objectnr)
              EndIf
            Else
              gs_rules()\inuse = #False
            EndIf
          Case #gs_plane                                                    ; it is a plane
            key = gs_objectkey(#gs_plane,objectnr)                            ; yes, so create a key
            If FindMapElement(gs_objects(),key)                             ; already in the list of objects
              gs_rules()\objectkey = key
              gs_rules()\inuse = #True
            Else                                                            ; not in the list of objects
              found = #False
              ResetList(gs_planes())
              While NextElement(gs_planes()) And found = #False
                If gs_planes()\plane = gs_rules()\objectnr
                  found = #True
                EndIf
              Wend
              If found = #True                                              ; is it known as a plane?
                gs_rules()\inuse = #True
                AddMapElement(gs_objects(),key)
                gs_rules()\objectkey = key
                gs_rules()\inuse = #True
                \plane = gs_planes()\plane
                \objecttype = #gs_plane
                \objectnr = objectnr
                \x = gs_planes()\x
                \y = gs_planes()\y
                \width = gs_planes()\width
                \height = gs_planes()\height
              Else
                gs_rules()\inuse = #False                                   ; unknown plane so mark as not in use
              EndIf
            EndIf
          Default
            gs_rules()\inuse = #False
            gs_debug("unknown object type")
          EndSelect
          ;
          ; find information on references for each rule
          ;
          If gs_rules()\inuse                                                ; only do a lookup for the reference if the object above was found
            referencenr = gs_rules()\referencenr
            Select gs_rules()\referencetype
            Case #gs_plane                                                   ; it is a plane
              key = gs_objectkey(#gs_plane,referencenr)
              If FindMapElement(gs_objects(),key)
                gs_rules()\referencekey = key                                ; plane already exists in the list of objects
                gs_rules()\inuse = #True
              Else
                AddMapElement(gs_objects(),key)                              ; plane did not yet exist in the list of objects
                gs_rules()\referencekey = key                                ; store the key to find back the object
                gs_rules()\inuse = #True
                \plane = gs_rules()\plane
                \objecttype = #gs_plane
                \objectnr = referencenr
                \x = GadgetX(referencenr)
                \y = GadgetY(referencenr)
                \width = GadgetWidth(referencenr)
                \height = GadgetHeight(referencenr)
              EndIf
            Case #gs_gadget                                                  ; it is a a gadget
              If IsGadget(referencenr)                                       ; does the gadget exist?
                key = gs_objectkey(#gs_gadget,referencenr)                     ; yes, so create a key
                If FindMapElement(gs_objects(),key)                          ; do we already know about it?
                  gs_rules()\referencekey = key                              ; yes, so store the key
                  gs_rules()\inuse = #True                                   ; flag the rule as valid
                Else                                                         ; we didn't find this gadget, let's add it
                  AddMapElement(gs_objects(),key)                            ; gadget exists, add gadget to table
                  gs_rules()\referencekey = key                              ; store the key to find back the object
                  gs_rules()\inuse = #True                                   ; flag the rule as valid
                  \plane = gs_rules()\plane
                  \objecttype = #gs_gadget
                  \objectnr = referencenr
                  \x = GadgetX(referencenr)
                  \y = GadgetY(referencenr)
                  \width = GadgetWidth(referencenr)
                  \height = GadgetHeight(referencenr)
                EndIf
              Else
                gs_rules()\inuse = #False                                    ; block the rule if the gadget did not exist
              EndIf
            Case #gs_window                                                  ; it is a window
              If IsWindow(referencenr)                                       ; does the window exist?
                key = gs_objectkey(#gs_window,referencenr)                     ; yes, so create a key
                If FindMapElement(gs_objects(),key)                          ; do we already know about it?
                  gs_rules()\referencekey = key                              ; yes, so store the key
                  gs_rules()\inuse = #True                                   ; flag the rule as valid
                Else                                                         ; we didn't find this gadget, let's add it
                  AddMapElement(gs_objects(),key)                            ; window exists, add window to table
                  gs_rules()\referencekey = key                              ; store the key to find back the object
                  gs_rules()\inuse = #True                                   ; flag the rule as valid
                  \plane = gs_rules()\plane
                  \objecttype = #gs_window
                  \objectnr = referencenr
                  \x = WindowX(referencenr)
                  \y = WindowY(referencenr)
                  \width = WindowWidth(referencenr)
                  \height = WindowHeight(referencenr)
                EndIf
              Else
                gs_rules()\inuse = #False                                    ; block the rule if the window did not exist
              EndIf
            Case #gs_element
              If ListSize(gs_elements()) > referencenr                       ; avoid errors
                key = gs_objectkey(#gs_element,referencenr)
                If FindMapElement(gs_objects(),key)                          ; do we already know about it?
                  gs_rules()\referencekey = key                              ; store the key
                  gs_rules()\inuse = #True                                   ; flag the rule as valid
                Else                                                         ; we didn't find this element, let's add it
                  AddMapElement(gs_objects(),key)                            ; element exists, add element to gs_objects() table
                  gs_rules()\referencekey = key                              ; store the key to find back the object
                  gs_rules()\inuse = #True
                  \plane = gs_rules()\plane
                  \objecttype = #gs_element
                  \objectnr = referencenr
                  SelectElement(gs_elements(),referencenr)
                  \x = gs_elements()\boundx
                  \y = gs_elements()\boundy
                  \width = gs_elements()\boundwidth
                  \height = gs_elements()\boundheight
                EndIf
              Else
                gs_rules()\inuse = #False
              EndIf
            Case #gs_null, #gs_none                                          ; not refering to anything, so refer to null
              key = gs_objectkey(#gs_none,#gs_none)
              gs_rules()\referencekey = key
            Default                                                          ; unknown type
              gs_debug("unknown reference type")
            EndSelect
          EndIf
          ;
          If gs_rules()\objectkey > "" And gs_rules()\objectkey = gs_rules()\referencekey                  ; filter out objects refering to themself
            gs_debug("rule refers to itself")
            gs_rules()\inuse = #False
          EndIf
          ;
        Else
          ;
          ; rule did not apply to plane selected when calling gs_apply(<plane>), mark it as not in use
          ;
          gs_rules()\inuse = #False
        EndIf
      EndWith
    Next
    ;
    ; step 5: store old coordinates so we can detect any changes to avoid unneccessary resizing or moving
    ;
    ForEach gs_objects()
      With gs_objects()
        \oldx = \x
        \oldy = \y
        \oldwith = \width
        \oldheight = \height
      EndWith
    Next
    ;
    ; step 6: process all rules in use
    ;
    ForEach gs_rules()
      With gs_objects()
        If gs_rules()\inuse = #True
          ;
          modifier1 = gs_rules()\modifier1
          modifier2 = gs_rules()\modifier2
          ;
          Select gs_rules()\rule
          Case #gs_none, #gs_parent, #gs_null
            FindMapElement(gs_objects(),gs_rules()\objectkey)
          Case #gs_setx, #gs_x1x1, #gs_alignleft
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x+modifier1
            If \x < modifier2
              \x = modifier2
            EndIf
          Case #gs_sety, #gs_y1y1, #gs_aligntop
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            y = \y
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \y = y+modifier1
            If \y < modifier2
              \y = modifier2
            EndIf
          Case #gs_setxy
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x
            y = \y
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x+modifier1
            \y = y+modifier2
          Case #gs_setwidth
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            width = \width
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \width = width+modifier1
            If \width < modifier2
              \width = modifier2
            EndIf
          Case #gs_setheight
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            height = \height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \height = height+modifier1
            If \height < modifier2
              \height = modifier2
            EndIf
          Case #gs_setsize
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            width = \width
            height = \height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \width = width+modifier1
            \height = height+modifier2
          Case #gs_use
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x
            y = \y
            width = \width
            height = \height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x+modifier1
            \y = y+modifier2
            \width = width
            \height = height
          Case #gs_stackup
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x
            y = \y
            width = \width
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x
            \y = y-\height-modifier1
            \width = width
          Case #gs_stackdown
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x
            y = \y+\height
            width = \width
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x
            \y = y+modifier1
            \width = width
          Case #gs_stackleft
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x
            y = \y
            height = \height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x-\width-modifier1
            \y = y
            \height = height
          Case #gs_stackright
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x
            y = \y
            width = \width
            height = \height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x+width+modifier1
            \y = y
            \height = height
          Case #gs_alignright, #gs_x2x2
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x+\width
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x-\width+modifier1
          Case #gs_alignbottom, #gs_y2y2
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            y = \y+\height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \y = y-\height+modifier1
          Case #gs_leftof, #gs_x2x1
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x-\width-modifier1
          Case #gs_rightof, #gs_x1x2
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \x+\width
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x+modifier1
          Case #gs_above, #gs_y2y1
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            y = \y
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \y = y-\height-modifier1
          Case #gs_below, #gs_y1y2
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            y = \y+\height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \y = y+modifier1
          Case #gs_stretchleft, #gs_x1x2
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x2 = \x+\width
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            x1 = x2+modifier1
            \width = \width+\x-x1
            \x = x1
            If \width < modifier2
              \width = modifier2
            EndIf
          Case #gs_stretchright, #gs_x2x1
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x1 = \x
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \width = x1-\x-modifier1
            If \width < modifier2
              \width = modifier2
            EndIf
          Case #gs_stretchup, #gs_y1y2
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            y2 = \y+\height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            y1 = y2+modifier1
            \height = \height+\y-y1
            \y = y1
            If \height < modifier2
              \height = modifier2
            EndIf
          Case #gs_stretchdown, #gs_y2y1
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            y1 = \y
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \height = y1-\y-modifier1
            If \height < modifier2
              \height = modifier2
            EndIf
          Case #gs_stretchparentleft
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \width = \x+\width-modifier1
            \x = modifier1
            If \width < modifier2
              \width = modifier2
            EndIf
          Case #gs_stretchparentright
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            width = \width
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \width = width-\x-modifier1
            If \width < modifier2
              \width = modifier2
            EndIf
          Case #gs_stretchparentbottom
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            height = \height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \height = height-\y-modifier1
            If \height < modifier2
              \height = modifier2
            EndIf
          Case #gs_alignparenttop
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \y = modifier1
          Case #gs_alignparentbottom
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            y = \height
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \y = y-\height-gs_rules()\modifier1
            If \y < modifier2
              \y = modifier2
            EndIf
          Case #gs_alignparentleft
            \x = modifier1
          Case #gs_alignparentright
            FindMapElement(gs_objects(),gs_rules()\referencekey)
            x = \width
            FindMapElement(gs_objects(),gs_rules()\objectkey)
            \x = x-\width-gs_rules()\modifier1
            If \x < modifier2
              \x = modifier2
            EndIf
          Default
            gs_debug("not yet implemented")
          EndSelect
          ;
          ; undo size changes for non-box elements
          ;
          If \objecttype = #gs_element
            If SelectElement(gs_elements(),\objectnr)
              Select gs_elements()\shape
              Case #gs_box, #gs_shadow
              Default
                \width = \oldwith
                \height = \oldheight
              EndSelect
            EndIf
          EndIf
        EndIf
      EndWith
    Next
    ;
    ; step 7: apply any changes to objects or their changeable properties
    ;
    ForEach gs_objects()
      With gs_objects()
        ;
        If \x <> \oldx Or \y <> \oldy Or \width <> \oldwith Or \height <> \oldheight
          ;
          ; protect against runaway conditions
          ;
          Select \objecttype
          Case #gs_gadget, #gs_window, #gs_image
            If \x < -99999 Or \x > 99999
              gs_debug("runaway condition x")
            EndIf
            If \y < -99999 Or \y > 99999
              gs_debug("runaway condition y")
            EndIf
            If \width < -99999 Or \width > 99999
              gs_debug("runaway condition width")
            EndIf
            If \height < -99999 Or \height > 99999
              gs_debug("runaway condition height")
            EndIf
          EndSelect
          ;
          ; resize and / or move the object
          ;
          Select \objecttype
          Case #gs_gadget
            ResizeGadget(\objectnr,\x,\y,\width,\height)
          Case #gs_window
            ResizeWindow(\objectnr,\x,\y,\width,\height)
          Case #gs_image
            If \width < 1
              \width = 1
            EndIf
            If \height < 1
              \height = 1
            EndIf
            CreateImage(\objectnr,\width,\height)
          Case #gs_element
            SelectElement(gs_elements(),\objectnr)
            If gs_elements()\shape = #gs_box
              If gs_elements()\handlex = gs_elements()\boundx
                gs_elements()\handlex = \x
              Else
                gs_elements()\handlex = gs_elements()\boundx+\width-1
              EndIf
              If gs_elements()\handley = gs_elements()\boundy
                gs_elements()\handley = \y
              Else
                gs_elements()\handley = gs_elements()\boundy+\height-1
              EndIf
              gs_elements()\width = \width
              gs_elements()\height = \height
              gs_elements()\boundwidth = \width
              gs_elements()\boundheight = \height
            Else
              gs_elements()\handlex = gs_elements()\handlex+\x-\oldx
              gs_elements()\handley = gs_elements()\handley+\y-\oldy
            EndIf
            gs_elements()\x = \x
            gs_elements()\y = \y
            gs_elements()\boundx = \x
            gs_elements()\boundy = \y
          Default
            gs_debug("unexpected")
          EndSelect
          ;
          ; somewhere some object's attribute(s) have changed...
          ;
          changes = #True
          \oldx = \x
          \oldy = \y
          \oldwith = \width
          \oldheight = \height
          ;
        EndIf
      EndWith
    Next
    ;
    ; step 8: create shadow elements and refers
    ;
    If gs_autoshadows
      ForEach gs_objects()
        With gs_objects()
          If \objecttype = #gs_gadget
            gs_shadow(\plane,\objecttype,\objectnr)
          EndIf
        EndWith
      Next
    EndIf
    ;
    If gs_autorefers
      gs_autorefers(plane)
    EndIf
    ;
    ; step 9: autoscale and autopan
    ;
    If changes And gs_viewpanereferencetype <> #gs_none
      If gs_autoscale
        gs_setviewpane(gs_viewpanewidth,gs_viewpaneheight,gs_viewpanereferencetype,gs_viewpanereferencenr)
        gs_viewscale = gs_viewpanewidth*oldviewscale/oldviewpanewidth
        gs_viewportx = oldviewportxc-gs_viewpanewidth/gs_viewscale/2
        gs_viewporty = oldviewportyc-gs_viewpaneheight/gs_viewscale/2
      ElseIf gs_autopan
        gs_setviewpane(gs_viewpanewidth,gs_viewpaneheight,gs_viewpanereferencetype,gs_viewpanereferencenr)
        gs_viewportx = oldviewportxc-gs_viewpanewidth/gs_viewscale/2
        gs_viewporty = oldviewportyc-gs_viewpaneheight/gs_viewscale/2
      EndIf
    EndIf
    ;
    ;
  EndIf
  ;
  ProcedureReturn changes                                     
  ;
EndProcedure

; IDE Options = PureBasic 4.51 (Windows - x86)
; CursorPosition = 404
; FirstLine = 15
; Folding = AAAAAAAA5
; Markers = 1964
; EnableAsm
; EnableUnicode
; EnableThread
; EnableXP
; Executable = g_spot.exe
; CurrentDirectory = C:\software\purebasic\_projects\
; EnableCompileCount = 1986
; EnableBuildCount = 3
; EnableExeConstant