;
EnableExplicit
IncludeFile "..\x_lib\x_lib.pb"
x_init()
IncludeFile "..\g_spot\g_spot.pb"
gs_init()
;
Declare init()
Declare main()
;
init()
main()
; End

Procedure init()                                                       ; initialize all sweetspot arrays, structures etc.
  ;
  ; *** initialize all constants and variables, load (standard) gui images and icons
  ;
  #placeholder = -999  ; look for "placeholder" or #placeholder for spots in the code that need to be expanded when supporting more gadget types, object types etc.
  ;
  Enumeration
    ;
    ; *** windows
    ;
    #w_main_nr
    ;
    ; *** menus
    ;
    ; *** toolbars
    ;
    ; *** gadgets
    ;
    #g_splitter_nr
    #g_table_nr
    #g_editor_nr
    #g_commit_nr
    ;
    ; *** actions
    ;
    #f_none
    #f_commit
    #f_exit
    ;
    ; *** images
    ;
  EndEnumeration
  ;
  ; *** global variables related to windows, events, gadgets etc.
  ;
  Global w_main_h.i, w_main_x.i, w_main_y,i, w_main_width.i, w_main_height.i
  Global event.i, event_gadget.i, event_menu.i, event_type.i
  Global f_action.i
  ;
  ; *** images
  ;
  ;
  ; *** lists and structures
  ;
  ; each gadget, window, object, has a number of parameters such as x, y, etc. these are represented by this table
  ;
  ;   source.s()       - contains all source code
  ;   table.table()    - contains all elements manipulated by sweetspot, regardless of their nature
  ;   code.code()      - is a list presenting a 'purabasic code' style view on the table
  ;   syntax.syntax()  - contains all syntax defintions
  ;
  ;
  ; *** source structures and lists
  ;
  ; the source list is a large list that contains all source code
  ;
  Structure source
    code.s                               ; a line with source code
    type.i                               ; any value of the constants above
  EndStructure
  ;
  Global NewList source.source()
  ;
  ; each source line can be of a specific type
  ;
  Enumeration #PB_Compiler_EnumerationValue    ; source types
    #type_sweetspot                      ; sweetspot feature related
    #type_window                         ; source defines / selects a window
    #type_gadget                         ; source defines a gadget
    #type_element                        ; source defines an element (as used in g_spot, not yet implemented)
    #type_slave                          ; source contains a line supposed to belong to the last non-slave line
    #type_code                           ; specific marked code unrelated to an object
    #type_comment                        ; a comment line
    #type_rule                           ; if a rule is directly tied into a gadget it's a 'slave' line, otherwise it's a 'rule' line
    #type_unrecognized                   ; unknown (likely code)
    #type_undefined                      ; new empty entry
  EndEnumeration
  ;
  ; typically, this should result in the table filled with sequences, ie. for example
  ;
  ;   index source code       source type
  ;   ----- ----------------- -------------------
  ;   0     beginform()       administrative
  ;   1     openwindow()      window
  ;   2     gs_rule()         slave
  ;   3     ;                 comment
  ;   4     q = q+1           code
  ;   5     ;                 comment
  ;   6     buttongadget()    gadget
  ;   7     setgadgetstate()  slave
  ;   8     imagegadget()     gadget
  ;   9     closegadgetlist() window
  ;   10    endform()         administrative
  ;
  ;
  ; *** table structures and lists
  ;
  ; the table list is a list of elements which can be manipulated by sweetspot, it's the 'central station' to all
  ; sweetspot functions, ie. any function interacts with the table, not directly with the original code
  ;
  ;   source <-> table <-> any other funtion
  ;
  ; the fields structure is used to keep track of all aspects of each object
  ;
  Structure fields
    ;
    ; field             use                                                used in, note
    ; ----------------  -------------------------------------------------  ------------------------------------------------
    ;
    x.i                 ; x-coordinate                                     all
    y.i                 ; y-coordinate                                     all
    width.i             ; width                                            all
    height.i            ; height                                           buttongadget, frame3dgadget, hyperlinkgadget
    scrollareawidth.i   ;                                                  scrollareagadget
    scrollareaheight.i  ;                                                  scrollareagadget
    scrollstep.i        ;                                                  scrollareagadget
    titlewidth.i        ; width of first column                            listicongadget
    image_h.s           ; handle to bitmap, imageid()                      imagegadget
    text.s              ; text displayed                                   buttongadget
    path.s              ; folder+patterns                                  explorerlistgadget, explorercomobogadget, explorertreegadget
    date.i              ; numeric value representing date                  calendargadget
    ipaddress.i         ; to set an ip address                             ipaddressgadget (needs an additioanl setgadgetstate)
    minimum.i           ; minimal value                                    progressbargadget, spingadget
    maximum.i           ; maximal value                                    progressbargadget, spingadget
    pagelength.i        ; page size
    shortcut.i          ; shortcut                                         shortcutgadget
    gadget1.i           ; child gadget nr1
    gadget2.i           ; child gadget nr2
    flags.i             ; optional flags
    tooltip.s           ;
    ;
  EndStructure
  ;
  Structure table
    ;
    ; field             use                                                used in, note
    ; ----------------  -------------------------------------------------  ------------------------------------------------
    ;
    source.s            ; original source
    code.s              ; current source                                   either original or generated by sweetspot
    sourcelinenr.i      ; position in source                               line number in source (starting with 0!)
    objecttype.i        ; object type (window, gadget, element)            see #type_xxxx above
    objectsubtype.i     ; subtype of object                                using the possible return values of gadgettype()
    objectsyntax.i      ; syntax template                                  entry in syntax() containing field information
    objectname.s        ; object number                                    original object number, a variable name or constant
    objectnr.i          ; object number (window / gadget number)           object number as assigned by sweetspot
    sequencenr.i        ; position in drawing sequence
    status.i            ; status of object                                 see the #status_xxxxx constants below
    warning.i           ; possible problems                                see the #warning_xxxxx constants below
    ;
    new.fields          ; a structure containing current properties
    old.fields          ; a structure containing the original properties
    ;
  EndStructure
  ;
  Global NewList table.table()
  ;
  ; status of the table element in table.table compared with its associated source element in source.source
  ;
  Enumeration #PB_Compiler_EnumerationValue     ; values for table()\status
    #status_ok          ; no changes
    #status_changed     ; one or more properties were changed (compared to the original source)
    #status_new         ; table entry did not exist in source
    #status_deleted     ; table entry removed
    #status_undefined   ; don't know yet (shouldn't happen though...)
  EndEnumeration
  ;
  Enumeration #PB_Compiler_EnumerationValue     ; values for table()\warning
    #warning_none       ; no problems detected
    #warning_other      ; something is wrong with this entry
  EndEnumeration
  ;
  ; *** code structures and lists
  ;
  ; the code list is just another view on the table, it contains a purebasic style representation of the elements in the table
  ; this list changes depending on the contents of table, whereas the source list doesn't change until committed from sweetspot
  ;
  Structure code
    code.s
  EndStructure
  ;
  Global NewList code.code()
  ;
  ;
  ; *** syntax definitions
  ;
  ; syntax definitions tell codecaddy how to deal with each line, command, and command parameters
  ;
  Structure syntaxfield
    ;
    ; field             use
    ; --------------------------  -------------------------------------------------
    ;
    tablefieldname.s              ; name of the field
    required.i                    ; one of the constants defined below
    preset.s                      ; preset (default) value
    ;
  EndStructure
  ;
  Enumeration #PB_Compiler_EnumerationValue     ; values for the syntaxfield\required
    ;
    #syntax_required              ; required field, throw an error if something is missing
    #syntax_optional              ; mark the field as optional
    #syntax_preset                ; assign a default value to this field
    ;
  EndEnumeration
  ;
  #syntax_fieldmax = 20                         ; max number of fields for syntax definitions / command parameters
  ;
  Structure syntax
    ;
    ; field                                     use
    ; ----------------------------------------  -------------------------------------------------
    ;
    definition.s                                ; original definition
    objecttype.i                                ; type of object, see above for the #type_xxxxx definitions
    objectsubtype.i                             ; for gadgets: the type of gadget, using gadgettype() values, -1 if unknown
    keyword.s                                   ; the purebasic command / procedure name used
    status.i                                    ; indicates a problem with the syntax definition
    syntaxfield.syntaxfield[#syntax_fieldmax]   ; how to interprete each field
    ;
  EndStructure
  ;
  Global NewList syntax.syntax()
  ;
  Enumeration #PB_Compiler_EnumerationValue     ; values for syntax()\status
    ;
    #syntax_ok
    #syntax_error
    ;
  EndEnumeration
  
  ; the syntax follows the regular purebasic syntax definitions, but is translated internally into something
  ; which is easier to process
  ;
EndProcedure

Procedure.i string_to_syntaxentry(string.s)                            ; create a syntax template to further process code
  Protected x.s, retval.i, optional.i, field_p.i
  ;
  ; *** add given line to the syntax table
  ;
  ; in:      string.s    - a syntax definition, in the common purebasic notation keyword ( <parameter> [ , <parameter> ] )
  ; retval:   -1            - error, incorrect syntax definition
  ;           0             - syntax definition accepted
  ;
  retval = 0
  AddElement(syntax())
  With syntax()
    ;
    ; store the definition
    ;
    \definition = string
    ;
    ; get the type of syntax
    ;
    x_parse(string," ",#x_parse_lcase|#x_parse_eol|#x_parse_trim|#x_parse_chars)
    Select x_parsed
    Case "gadget"
      \objecttype = #type_gadget
    Case "window"
      \objecttype = #type_window
    Default
      \objecttype = -1
      retval = -1
      x_debug("unknown object type")
    EndSelect
    ;
    ; get the keyword
    ;
    x_parsestring("(")
    \keyword = x_parsed
    ;
    ; get the subtype
    ;
    Select x_parsed
    Case "openwindow"
      \objectsubtype = -1
    Case "buttongadget"
      \objectsubtype = #PB_GadgetType_Button
    Case "imagegadget"
      \objectsubtype = #PB_GadgetType_Image
    Case "placeholder"
      ;
      ; <-- add new / missing constants /subtypes here
      ;
    Default
      \objectsubtype = -1
      x_debug("unknown object subtype")
    EndSelect
    ;
    ;
    ; run through all parameters, do this by repeatedly looking for , [ or )
    ;
    ;   , indicates next parameter field
    ;   [ toggles 'optional parameters'
    ;   ] end of optional parameters, simply ignore
    ;   ) indicates we're done
    ;
    optional.i = #False
    field_p.i = 0
    ;
    While retval > -1 And x_parse > "" And field_p < #syntax_fieldmax
      If x_parse_match = "["
        optional = #True
        x_parsestring(",")
        If x_parsed <> ""
          retval = -1
        EndIf
      EndIf
      If x_parsestring("[]),")
        If x_parse_match <> ")"
          ;
          ; if everything went okay we should end up here with a proper field definition
          ; filter out all unrecognized field names
          ;
          Select x_parsed
          Case "objectnr"
          Case "x", "y", "width", "height"
          Case "scrollareawidth", "scrollareaheight", "scrollstep"
          Case "titlewidth", "image_h", "text", "path", "date", "ipaddress"
          Case "minimum", "maximum", "pagelength", "shortcut", "gadget1", "gadget2"
          Case "flags", "tooltop"
          Default
            x_debug("unknown field name '"+x_parsed+"'")
            retval = -1
          EndSelect
          ;
          ; fill in the fields
          ;
          \syntaxfield[field_p]\tablefieldname = x_parsed
          If optional
            \syntaxfield[field_p]\required = #syntax_optional
          Else
            \syntaxfield[field_p]\required = #syntax_required
          EndIf
          field_p = field_p+1
          ;
          ; all this begs the question if i should grey out unused fields in the table element properties lateron, as certain
          ; fields are not used for certain commands... sigh... I guess I have to
          ;
        EndIf
      EndIf
    Wend
    ;
    If retval = 0
      \status = #syntax_ok
    Else
      \status = #syntax_error
    EndIf
    ;
  EndWith
  ;
  ProcedureReturn retval
EndProcedure

Procedure cleartableentry(n.i)                                         ; clear specified entry in table
  SelectElement(table(),n)
  With table()
    \source = ""
    \code = ""
    \objectname = ""
    \objectnr = 0
    \objecttype = #type_undefined
    \objectsubtype = #PB_GadgetType_Unknown
    \objectsyntax = -1
    \sequencenr = 0
    \sourcelinenr = 0
    \status = #status_undefined
    \warning = #warning_none
  EndWith
  With table()\new
    \x.i                = 0
    \y.i                = 0
    \width.i            = 0
    \height.i           = 0
    \scrollareawidth.i  = 0
    \scrollareaheight.i = 0
    \scrollstep.i       = 0
    \titlewidth.i       = 0
    \image_h.s          = ""
    \text.s             = ""
    \path.s             = ""
    \date.i             = 0
    \ipaddress.i        = 0
    \minimum.i          = 0
    \maximum.i          = 0
    \pagelength.i       = 0
    \shortcut.i         = 0
    \gadget1.i          = 0
    \gadget2.i          = 0
    \flags.i            = 0
    \tooltip.s          = ""
  EndWith
  With table()\old
    \x.i                = 0
    \y.i                = 0
    \width.i            = 0
    \height.i           = 0
    \scrollareawidth.i  = 0
    \scrollareaheight.i = 0
    \scrollstep.i       = 0
    \titlewidth.i       = 0
    \image_h.s          = ""
    \text.s             = ""
    \path.s             = ""
    \date.i             = 0
    \ipaddress.i        = 0
    \minimum.i          = 0
    \maximum.i          = 0
    \pagelength.i       = 0
    \shortcut.i         = 0
    \gadget1.i          = 0
    \gadget2.i          = 0
    \flags.i            = 0
    \tooltip.s          = ""
  EndWith
  ;
EndProcedure







; need to redo this section
; just dunno yet how
; but each and every entry by hand seems to be the wrong way
;
Structure pbconstants
  objecttype.i
  objectsubtype.i
  constantname.s
  constantvalue.i
EndStructure
;
Global NewList pbconstants.pbconstants()
;
With pbconstants()
  AddElement(pbconstants())
  \objecttype = #type_gadget
  \objectsubtype = #PB_GadgetType_Button
  \constantname = "#PB_Button_Toggle"
  \constantvalue = #PB_Button_Toggle
EndWith


Procedure.i gadgetconstants_to_val(s.s)                                ; turn a string containing purebasic constants for gadgets into a numeric value
  Protected p.i, n.i, retval.i
  ;
  ; *** check the string and see if it contains any known (read: entered, sigh) purebasic constants that might make sense
  ;
  ; in:        s.s    - string to parse, containing one or more constants seperated by pipe characters
  ; retval:    -1     - did not recognize one or more parameters
  ;            n      - numeric value of all fields combined
  ;
  ; notes:
  ;
  ; - if the very first digit in the string is NOT a hash '#' symbol the string will be considered a direct numeric value either in hex or dec
  ; - this procedure needs to be extended every time new / missing flags come up, or purebasic features new as yet unknown flags
  ; - see also windowconstants_to_val()
  ;
  If Left(s,1) <> "#"
    retval = x_val(s)
  Else
    ;
    retval = 0
    n = CountString(s,"|")+1
    p = 0
    While p < n
      p = p+1
      Select StringField(s,p,"|")
      Case "#pb_button_toggle"
        retval = retval | #PB_Button_Toggle
      Case "placeholder"
        ;
        ; <-- add new / missing constants here
        ;
      Default
        retval = -1
        x_debug("unknown gadget 'flags' constant '"+StringField(s,p,"|")+"'")
      EndSelect
    Wend
    ;
  EndIf
  ;
  ProcedureReturn retval
EndProcedure

Procedure.i windowconstants_to_val(s.s)                                ; turn a string containing purebasic constants for windows into a numeric value
  Protected p.i, n.i, retval.i
  ;
  ; *** check the string and see if it contains any known (read: entered, sigh) purebasic constants that might make sense
  ;
  ; in:        s.s    - string to parse, containing one or more constants seperated by pipe characters
  ; retval:    -1     - did not recognize one or more parameters
  ;            n      - numeric value of all fields combined
  ;
  ; notes:
  ;
  ; - if the very first digit in the string is NOT a hash '#' symbol the string will be considered a direct numeric value either in hex or dec
  ; - this procedure needs to be extended every time new / missing flags come up, or purebasic features new as yet unknown flags
  ; - see also gadgetconstants_to_val()
  ;
  If Left(s,1) <> "#"
    retval = x_val(s)
  Else
    ;
    retval = 0
    n = CountString(s,"|")+1
    p = 0
    While p < n
      p = p+1
      Select StringField(s,p,"|")
      Case "#pb_window_screencentered"
        retval = retval | #PB_Window_ScreenCentered
      Case "#pb_window_systemmenu"
        retval = retval | #PB_Window_SystemMenu
      Case "#pb_window_sizegadget"
        retval = retval | #PB_Window_SizeGadget
      Case "placeholder"
        ;
        ; <-- add new / missing constants here
        ;
      Default
        retval = -1
        x_debug("unknown window 'flags' constant '"+StringField(s,p,"|")+"'")
      EndSelect
    Wend
    ;
  EndIf
  ;
  ProcedureReturn retval
EndProcedure

Procedure.s val_to_windowconstants(n.i)                                ; turn a value into a string with purebasic constants for windows
  Protected retval.s
  ;
  If n & #PB_Window_ScreenCentered
    n = n - #PB_Window_ScreenCentered
  EndIf
  If n & #PB_Window_SizeGadget
  EndIf
  If n & #PB_Window_SystemMenu
  EndIf
  
  ProcedureReturn retval
EndProcedure

Procedure.s val_to_gadgetconstants(n.i)                                ; turn a value into a string with purebasic constants for gadgets
  Protected retval.s
  ;
  ProcedureReturn retval
EndProcedure











Procedure.i codeline_to_tableentry(code.s,n.i,insert.i=#False,sourcelinenr=-1)    ; convert a line of purebasic code into a table entry
  Protected retval.i, syntax_p.i, field_p.i, v.i
  ;
  ; *** import a single line into the table, insert or replace at specified position
  ;
  ; in:      n.i = -1             ; add to the end of the table() list
  ;                n              ; replace or insert at line n
  ;          insert.i             ; set to #true to insert at specified position, otherwise it replaces existing entry
  ;          sourcelinenr.i = -1  ; do not replace the original source line or linenumber
  ;                         = n   ; replace table()\source.s with element <sourcelinenr> from the source() list
  ; retval:  0                    ; no problems
  ;          -1                   ; could not process line properly
  ;
  ; notes:
  ;
  ; - this is the inverse of tableentry_to_codeline()
  ;
  ;
  ; first find some space for the table entry
  ;
  retval = 0
  If n = -1
    LastElement(table())
  Else
    SelectElement(table(),n)
  EndIf
  If insert = #True Or n = -1
    AddElement(table())
  Else
    cleartableentry(n)
  EndIf
  ;
  With table()
    ;
    ; store original source if required
    ;
    If sourcelinenr > -1
      SelectElement(source(),sourcelinenr)
      \source = source()\code
      \sourcelinenr = sourcelinenr
    EndIf
    \code = code
    ;
    ; look up the keyword in the syntax() list
    ;
    x_parse(code,"(",#x_parse_lcase|#x_parse_trim|#x_parse_eol|#x_parse_chars)
    retval = -1
    syntax_p = 0
    ResetList(syntax())
    Repeat
      NextElement(syntax())
      If syntax()\keyword = x_parsed
        retval = 0
      Else
        syntax_p = syntax_p+1
      EndIf
    Until retval = 0 Or syntax_p = ListSize(syntax())
    ;
    If retval = 0
      ;
      ; found a valid entry (or so we assume :-))
      ;
      ; store type (either a window or a gadget)
      ;
      \objecttype = syntax()\objecttype          ; store object type, if known
      \objectsubtype = syntax()\objectsubtype    ; store object subtype, if known or applicable
      \objectsyntax = syntax_p                   ; store a pointer into the syntax table to speed up lookup activities
      ;
      field_p = 0
      While x_parsed > ""
        x_parsestring(",)")
        v = Val(x_parsed)
        If x_parsed > ""
          ;
          ; parameter found, store it in the appropriate field in table()
          ;
          Select syntax()\syntaxfield[field_p]\tablefieldname
          Case "objectnr"
            \objectname = x_parsed
          Case "x"
            \new\x = v
          Case "y"
            \new\y = v
          Case "width"
            \new\width = v
          Case "height"
            \new\height = v
          Case "text"
            \new\text = x_parsed
          Case "image_h"
            \new\image_h = x_parsed
          Case "flags"
            ;
            ; this is pretty sad as i have to reverse the flag values to their numeric counterparts, sigh
            ; and in the properties gadget i'll have to redo this again, double sigh
            ;
            ; how flags are actually interpreted depends on the recognized table()\objecttype and table()\objectsubtype
            ; ie. window or gadget+gadgettype
            ;
            Select \objecttype
            Case #type_window
              ;
              ; interprete the flags parameter for openwindow()
              ;
              \new\flags = windowconstants_to_val(x_parsed)
              ;
            Case #type_gadget
              ;
              ; interprete the flags parameter for gadgets
              ;
              Select \objectsubtype
              Case #PB_GadgetType_Button, #PB_GadgetType_Image
                \new\flags = gadgetconstants_to_val(x_parsed)
              Case #placeholder
                ;
                ; <-- add new / additional types here
                ;
              Default
                x_debug("not yet implemented or unknown object subtype")
              EndSelect
              ;
            Default
              x_debug("not yet implemented or unknown object type")
            EndSelect
            ;
          Default
            x_debug("unknown field name '"+syntax()\syntaxfield[field_p]\tablefieldname+"'")
          EndSelect
          ;
        EndIf
        field_p = field_p+1
      Wend
    EndIf
  EndWith
  ;
  ProcedureReturn retval
EndProcedure

Procedure.s tableentry_to_codeline(n.i)                                           ; generate a line of purebasic code using the specified table entry
  Protected retval.s, syntax_p.i, field_p.i, fieldname.s
  ;
  ; notes:
  ;
  ; - this is the inverse of codeline_to_tableentry()
  ;
  With table()
    SelectElement(table(),n)
    SelectElement(syntax(),\objectsyntax)
    ;
    retval = syntax()\keyword+"("
    ;
    field_p = 0
    Repeat
      fieldname = syntax()\syntaxfield[field_p]\tablefieldname
      Select fieldname
      Case ""
        ; no more entries
      Case "objectnr"
        retval = retval + \objectname + ","
      Case "x"
        retval = retval + Str(\new\x) + "."
      Case "y"
        retval = retval + Str(\new\y) + "."
      Case "width"
        retval = retval + Str(\new\width) + "."
      Case "height"
        retval = retval + Str(\new\height) + "."
      Case "text"
        retval = retval + \new\text + ","
      Case "image_h"
        retval = retval + \new\image_h
      Case "placeholder"
        ;
        ; <-- add new field names here
        ;
      Case "flags"
        ;
        Select \objecttype
        Case #type_window
          
        Case #type_gadget
        Case #placeholder
          ;
          ; <-- add new object types here
          ;
        Default
          x_debug("not yet implemented or unknown object type")
        EndSelect
      Default
        x_debug("unknown field name '"+fieldname+"'")
      EndSelect
      ;
      field_p = field_p+1
    Until field_p = #syntax_fieldmax Or fieldname = ""
  EndWith
  ;
  ProcedureReturn retval
EndProcedure

Debug string_to_syntaxentry("window Openwindow( objectnr , x , y , width , height , text [ , flags ] )")
Debug string_to_syntaxentry("gadget ButtonGadget( objectnr , x , y , width , height , text [ , flags ] )")
Debug string_to_syntaxentry("gadget ImageGadget( objectnr , x , y , width , height , image_h [ , flags ] )")
Debug codeline_to_tableentry("OpenWindow(1,100,100,200,200,"+#DQUOTE$+"window"+#DQUOTE$+",#PB_Window_ScreenCentered|#PB_Window_SystemMenu)",-1)
Debug codeline_to_tableentry("ButtonGadget(#g_1,10,10,30,20,"+#DQUOTE$+"A"+#DQUOTE$+",#PB_Button_Toggle)",-1)
Debug codeline_to_tableentry("ButtonGadget(#g_2,50,10,30,20,"+#DQUOTE$+"B"+#DQUOTE$+",#PB_Button_Toggle)",-1)
Debug codeline_to_tableentry("ImageGadget(#g_3,10,40,80,40,#i_image1",-1)
Debug tableentry_to_codeline(0)
;
; so all this bull only hes created a little table yet, sigh
; and i have not even checked if the table is properly filled, double sigh
; next two steps are probalby test draw something on screen depending on the contents of the table, and export it back to code (roundtrip, yeah!)

Procedure.i callback_handler(windowid.l,message.l,wparam.l,lparam.l)
  Protected *nmhdr.NMHDR
  Protected *nmlvcustomdraw.NMLVCUSTOMDRAW
  Protected gadget_h.l, item_nr.l
  ;
  Select message
  Case #WM_NOTIFY
    *nmlvcustomdraw = lparam                              ; use to do custom drawing
    *nmhdr.NMHDR = lparam                                 ; use to grab tab key
    gadget_h = *nmhdr\hwndFrom                            ; handle of gadget that caused event
    If GetAsyncKeyState_(#VK_TAB) = 32768                 ; this is all about processing tab keys etc.
      Select gadget_h                                     ; we can process multiple gadgets upon a tab
      Case GadgetID(#g_editor_nr)                         ; tab was pressed on gadget number 'x'
        SetActiveGadget(#g_commit_nr)                     ; or whatever is the next gadget
        ProcedureReturn 1                                 ; don't do anything else with this message
      EndSelect
    EndIf
  EndSelect
  ProcedureReturn #PB_ProcessPureBasicEvents              ; continue processing messages
EndProcedure

Procedure main()
  ;
  ; *** create window
  ;
  w_main_h = OpenWindow(#w_main_nr,10,10,400,300,"Test",#PB_Window_ScreenCentered|#PB_Window_SystemMenu|#PB_Window_SizeGadget)
  w_main_width = WindowWidth(#w_main_nr)
  w_main_height = WindowHeight(#w_main_nr)
  w_main_x = WindowX(#w_main_nr)
  w_main_y = WindowY(#w_main_nr)
  SetWindowCallback(@callback_handler())
  ;
  ; *** add a menu
  ;
  ; *** add toolbar and toolbarbuttons
  ;
  ; *** create gadgets
  ;
  TreeGadget(#g_table_nr,10,10,150,20)
  x_editorgadget(#g_editor_nr,80,10,60,20,"","",#x_gadget_notify|#x_gadget_notify)
  SplitterGadget(#g_splitter_nr,10,10,10,10,#g_table_nr,#g_editor_nr)
  gs_rule(0,#gs_setsize,#gs_gadget,#g_splitter_nr,#gs_window,#w_main_nr,-20,-50)
  ;
  ButtonGadget(#g_commit_nr,10,10,60,20,"Commit")
  gs_rule(0,#gs_alignparentright,#gs_gadget,#g_commit_nr,#gs_window,#w_main_nr,10,210)
  gs_rule(0,#gs_alignparentbottom,#gs_gadget,#g_commit_nr,#gs_window,#w_main_nr,10,120)
  ;
  ; *** event loop
  ;
  ; Repeat
  event = WaitWindowEvent()
  event_gadget = EventGadget()
  event_menu = EventMenu()
  event_type = EventType()
  ;
  ; *** event processing
  ;
  f_action = #f_none
  Select event
  Case #PB_Event_Menu
  Case #PB_Event_Gadget
    Select event_gadget
    Case #g_commit_nr
      f_action = #f_commit
    EndSelect
  Case #PB_Event_MoveWindow, #PB_Event_SizeWindow
    If gs_apply()
    EndIf
  Case #PB_Event_CloseWindow
    f_action = #f_exit
  Default
  EndSelect
  ;
  ; *** action processing
  ;
  Select f_action
  Case #f_none
  Case #f_commit
  Case #f_exit
  EndSelect
  ;  Until f_action = #f_exit
  ;
EndProcedure


; IDE Options = PureBasic 4.41 (Windows - x86)
; CursorPosition = 444
; FirstLine = 3
; Folding = AAIL-
; Markers = 118,370
; EnableAsm
; EnableUnicode
; EnableThread
; EnableXP
; CurrentDirectory = C:\software\purebasic\_projects\
; EnableCompileCount = 250
; EnableBuildCount = 0
; EnableExeConstant