﻿; x_lib_winsock ejn (blueznl)
; testfile for winsock functions in x_lib
; bluez@xs4all.nl
;
; note: run client and server both in unicode or in asci mode
;
; most of the code of the old ws_lib is no longer required, as the definitions have been included in purebasic
; the few lines that were still valid have been moved into x_lib() though the code involved is so small
; it could be easily inserted in here (a few structures and a few minimal functions)
;
IncludeFile "x_lib.pb"
;
x_init()                                                ; some general stuff
x_ws_start()                                            ; load dll and start winsock
;
InitKeyboard()                                          ; to enable keyboard commands and constants
OpenConsole()                                           ; console for input / output
;

host_name.s = x_ws_hostname()                           ; get hostname
host_ip.s = x_ws_nametoip(host_name)                    ; get ip address
;
PrintN("this is "+host_name+" at "+host_ip)
;
Repeat
  PrintN("[c] client [s] server [x] exit")
  Repeat
    choice.s = x_waitkey()
    If choice = Chr(27)
      choice = "X"
    EndIf
  Until choice>""
  Select choice
  Case "S"
    ;
    ; open socket for tcp/ip
    ;
    PrintN("")
    PrintN("SERVER")
    PrintN("")
    ;
    server_socket.l = SOCKET_(#AF_INET,#SOCK_STREAM,#IPPROTO_TCP)
    PrintN("server socket "+Str(server_socket))
    ;
    server_port.l = 6000
    server_bport.l = htons_(server_port)                        ; port number to tcp/ip network byte order
    PrintN("server port "+Str(server_port))
    ;
    ; bind a socket (reserve socket and port for exclusive use by the application)
    ;
    server_sockaddr.x_ws_sockaddr

    server_sockaddr\in_family = #AF_INET                        ; udp / tcp
    server_sockaddr\in_port = server_bport                      ; port
    retval = bind_(server_socket,@server_sockaddr,16)           ; and bind it
    PrintN("bind "+Str(retval))
    ;
    ; set socket usage to nonblocking
    ;
    ; use either wsaasyncselect which allows windows event messages:
    ;
    ;   retval = WSAAsyncSelect_(socket,window_handle,message,#FD_READ|#FD_ACCEPT|#FD_CONNECT|#FD_CLOSE)
    ;
    ; or use ioctlsocket_():
    ;
    server_mode.l = 1                                           ; 1 = non-blocking, 0 = blocking
    retval = ioctlsocket_(server_socket,#FIONBIO,@server_mode)  ; 0 = no error
    PrintN("set (non)blocking mode "+Str(retval))
    ;
    ; listen for incoming calls on bound port, set queue size (1..5, here 4)
    ;
    retval = listen_(server_socket,4)
    PrintN("listen "+Str(retval))
    ;
    PrintN("waiting for clients (any key to abort)")
    ;
    Repeat
      ;
      ; accept incoming connections and assign them to new sockets
      ; this is a blocking function if not set to non-blocking
      ;
      ; if you don't care about the callers identity:
      ;
      ;   retval = accept_(server_socket,0,0)
      ;
      ; if you do care, you can retrieve some info from a struct:
      ;
      accept_sockaddr.x_ws_sockaddr                 ; structure for more info
      accept_sockaddr_l=16                          ; length of the structure
      accept_socket.l = accept_(server_socket,@accept_sockaddr,@accept_sockaddr_l)
      ;
      If accept_socket > 0
        ;
        ; there was a call and a socket was created to handle it
        ;
        PrintN("connection accepted on socket "+Str(accept_socket))
        caller_bip.l = accept_sockaddr\in_addr
        caller_ip.s = x_ws_biptoip(caller_bip)
        caller_name.s = x_ws_biptoname(caller_bip)
        PrintN("from "+caller_name+" at "+caller_ip)
        ;
        ; note on socket options: SO_DONTLINGER is set by default
        ; it's possible to enable a 'heartbeat' on a socket / connection
        ;
        ; socket_keepalive.b=1
        ; retval = setsockopt_(accept_socket,#SOL_SOCKET,#SO_KEEPALIVE,@socket_keep_alive,1)
        ; PrintN("socket option: keep alive "+STR(retval))
        ;
        done.l = #False      ; if done then close socket, either gracefull or due to an error
        Repeat
          ;
          ; check if something can be received
          ;
          ; status gives the number of sockets that have data
          ; if a socket is gracefully closed select_() will also report that recv_() can be used
          ;
          ; timeout.ws_timeval
          ; timeout\ws_seconds=0
          ; timeout\ws_microseconds=1
          ; accept_fdset.ws_fd_set
          ; accept_fdset\ws_fd_number = 1                     ; number of entries in the set
          ; accept_fdset\ws_fd_socket[1] = accept_socket      ; one word per socket
          ; status.l=select_(0,@accept_fdset,0,0,@timeout)
          ;
          ; receive data
          ;
          buffer_l.l = 1024       ; buffer of 1024 kbytes
          buffer_p = AllocateMemory(buffer_l)
          received.l = recv_(accept_socket,buffer_p,buffer_l,0)
          ;
          ; received>0   number of bytes received aka transferred to buffer
          ; received=0   gracefull disconnect (properly closed by other side)
          ; received<0   error, or no data
          ;
          If received > 0                           ; data to receive
            Print(PeekS(buffer_p,buffer_l))
          ElseIf received = 0                       ; gracefull disconnect
            done = #True                            ; so close the socket
          Else
            error.l = WSAGetLastError_()
            If error=#WSAEWOULDBLOCK                ; when non-blocking: no more data to read
              ;
              ; PrintN("Error #WSAEWOULDBLOCK "+Str(error))
              ;
              Delay(100)
            Else                                    ; otherwise a real problem
              PrintN("Error "+Str(error))
              close = 1                             ; so close the socket
            EndIf
          EndIf
          ;
          FreeMemory(buffer_p)                    ; free the read buffer
          ;
        Until done = #True
        ;
        closesocket_(accept_socket)                 ; close socket
        PrintN("")
        ;
      Else
        ;
        ; no incoming calls, so let's free up some cpu resources
        ;
        Delay(100)
      EndIf
      ;
    Until Inkey()>""
    closesocket_(server_socket)
    PrintN("")
    ;
  Case "C"
    ;
    PrintN("CLIENT")
    PrintN("")
    ;
    server_ip.s = host_ip
    server_bip.l = x_ws_iptobip(server_ip)
    server_port.l = 6000
    server_bport.l = htons_(server_port)
    ;
    PrintN("trying to connect to "+server_ip+" port "+Str(server_port))
    PrintN("(any key to abort)")
    ;
    n.l = 0
    ;
    Repeat
      ;
      n = n+1
      PrintN("attempt "+Str(n))
      ;
      client_socket.l = SOCKET_(#AF_INET,#SOCK_STREAM,#IPPROTO_TCP)
      PrintN("client socket "+Str(client_socket))
      ;
      ; *** connect to another server / port using an empty socket
      ;
      client_sockaddr.x_ws_sockaddr
      client_sockaddr\in_family = #AF_INET
      client_sockaddr\in_port = server_bport
      client_sockaddr\in_addr = server_bip
      client_sockaddr_l.l = 16
      retval = connect_(client_socket,@client_sockaddr,client_sockaddr_l)
      ;
      PrintN("connect "+Str(retval))
      ;
      message.s = "call " +Str(n)+" from "+host_name+" at "+host_ip
      send.l = send_(client_socket,@message,Len(message),0)
      ;
      closesocket_(client_socket)
      ;
    Until Inkey()>""
    PrintN("")
  EndSelect
Until choice="X"
;
x_ws_stop()
x_end()

; IDE Options = PureBasic 4.51 (Windows - x86)
; CursorPosition = 9
; Markers = 19
; EnableAsm
; EnableUnicode
; EnableThread
; EnableXP
; CurrentDirectory = C:\software\purebasic\_projects\
; EnableCompileCount = 15
; EnableBuildCount = 0
; EnableExeConstant